import duckdb
import pytest
import threading
import queue as Queue
import numpy as np
from conftest import NumpyPandas, ArrowPandas
import os
from typing import List

try:
    import pyarrow as pa

    can_run = True
except ImportError:
    can_run = False


def connect_duck(duckdb_conn):
    out = duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)').fetchall()
    assert out == [(42,), (84,), (None,), (128,)]


def everything_succeeded(results: List[bool]):
    return all([result == True for result in results])


class DuckDBThreaded:
    def __init__(self, duckdb_insert_thread_count, thread_function, pandas):
        self.duckdb_insert_thread_count = duckdb_insert_thread_count
        self.threads = []
        self.thread_function = thread_function
        self.pandas = pandas

    def multithread_test(self, result_verification=everything_succeeded):
        duckdb_conn = duckdb.connect()
        queue = Queue.Queue()

        # Create all threads
        for i in range(0, self.duckdb_insert_thread_count):
            self.threads.append(
                threading.Thread(
                    target=self.thread_function, args=(duckdb_conn, queue, self.pandas), name='duckdb_thread_' + str(i)
                )
            )

        # Record for every thread if they succeeded or not
        thread_results = []
        for i in range(0, len(self.threads)):
            self.threads[i].start()
            thread_result: bool = queue.get(timeout=60)
            thread_results.append(thread_result)

        # Finish all threads
        for i in range(0, len(self.threads)):
            self.threads[i].join()

        # Assert that the results are what we expected
        assert result_verification(thread_results)


def execute_query_same_connection(duckdb_conn, queue, pandas):
    try:
        out = duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)')
        queue.put(False)
    except:
        queue.put(True)


def execute_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)')
        queue.put(True)
    except:
        queue.put(False)


def insert_runtime_error(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.execute('insert into T values (42), (84), (NULL), (128)')
        queue.put(False)
    except:
        queue.put(True)


def execute_many_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        # from python docs
        duckdb_conn.execute(
            """
                CREATE TABLE stocks(
                    date text,
                    trans text,
                    symbol text,
                    qty real,
                    price real
                )
            """
        )
        # Larger example that inserts many records at a time
        purchases = [
            ('2006-03-28', 'BUY', 'IBM', 1000, 45.00),
            ('2006-04-05', 'BUY', 'MSFT', 1000, 72.00),
            ('2006-04-06', 'SELL', 'IBM', 500, 53.00),
        ]
        duckdb_conn.executemany(
            """
            INSERT INTO stocks VALUES (?,?,?,?,?)
        """,
            purchases,
        )
        queue.put(True)
    except:
        queue.put(False)


def fetchone_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)').fetchone()
        queue.put(True)
    except:
        queue.put(False)


def fetchall_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)').fetchall()
        queue.put(True)
    except:
        queue.put(False)


def conn_close(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.close()
        queue.put(True)
    except:
        queue.put(False)


def fetchnp_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)').fetchnumpy()
        queue.put(True)
    except:
        queue.put(False)


def fetchdf_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)').fetchdf()
        queue.put(True)
    except:
        queue.put(False)


def fetchdf_chunk_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)').fetch_df_chunk()
        queue.put(True)
    except:
        queue.put(False)


def fetch_arrow_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)').fetch_arrow_table()
        queue.put(True)
    except:
        queue.put(False)


def fetch_record_batch_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        duckdb_conn.execute('select i from (values (42), (84), (NULL), (128)) tbl(i)').fetch_record_batch()
        queue.put(True)
    except:
        queue.put(False)


def transaction_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    duckdb_conn.execute("CREATE TABLE T ( i INTEGER)")
    try:
        duckdb_conn.begin()
        duckdb_conn.execute('insert into T values (42), (84), (NULL), (128)')
        duckdb_conn.rollback()
        duckdb_conn.execute('insert into T values (42), (84), (NULL), (128)')
        duckdb_conn.commit()
        queue.put(True)
    except:
        queue.put(False)


def df_append(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    duckdb_conn.execute("CREATE TABLE T ( i INTEGER)")
    df = pandas.DataFrame(np.random.randint(0, 100, size=15), columns=['A'])
    try:
        duckdb_conn.append('T', df)
        queue.put(True)
    except:
        queue.put(False)


def df_register(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    df = pandas.DataFrame(np.random.randint(0, 100, size=15), columns=['A'])
    try:
        duckdb_conn.register('T', df)
        queue.put(True)
    except:
        queue.put(False)


def df_unregister(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    df = pandas.DataFrame(np.random.randint(0, 100, size=15), columns=['A'])
    try:
        duckdb_conn.register('T', df)
        duckdb_conn.unregister('T')
        queue.put(True)
    except:
        queue.put(False)


def arrow_register_unregister(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    arrow_tbl = pa.Table.from_pydict({'my_column': pa.array([1, 2, 3, 4, 5], type=pa.int64())})
    try:
        duckdb_conn.register('T', arrow_tbl)
        duckdb_conn.unregister('T')
        queue.put(True)
    except:
        queue.put(False)


def table(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    duckdb_conn.execute("CREATE TABLE T ( i INTEGER)")
    try:
        out = duckdb_conn.table('T')
        queue.put(True)
    except:
        queue.put(False)


def view(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    duckdb_conn.execute("CREATE TABLE T ( i INTEGER)")
    duckdb_conn.execute("CREATE VIEW V as (SELECT * FROM T)")
    try:
        out = duckdb_conn.values([5, 'five'])
        queue.put(True)
    except:
        queue.put(False)


def values(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        out = duckdb_conn.values([5, 'five'])
        queue.put(True)
    except:
        queue.put(False)


def from_query(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    try:
        out = duckdb_conn.from_query("select i from (values (42), (84), (NULL), (128)) tbl(i)")
        queue.put(True)
    except:
        queue.put(False)


def from_df(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    df = pandas.DataFrame(['bla', 'blabla'] * 10, columns=['A'])
    try:
        out = duckdb_conn.execute("select * from df").fetchall()
        queue.put(True)
    except:
        queue.put(False)


def from_arrow(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    arrow_tbl = pa.Table.from_pydict({'my_column': pa.array([1, 2, 3, 4, 5], type=pa.int64())})
    try:
        out = duckdb_conn.from_arrow(arrow_tbl)
        queue.put(True)
    except:
        queue.put(False)


def from_csv_auto(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'integers.csv')
    try:
        out = duckdb_conn.from_csv_auto(filename)
        queue.put(True)
    except:
        queue.put(False)


def from_parquet(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'binary_string.parquet')
    try:
        out = duckdb_conn.from_parquet(filename)
        queue.put(True)
    except:
        queue.put(False)


def description(duckdb_conn, queue, pandas):
    # Get a new connection
    duckdb_conn = duckdb.connect()
    duckdb_conn.execute('CREATE TABLE test (i bool, j TIME, k VARCHAR)')
    duckdb_conn.execute("INSERT INTO test VALUES (TRUE, '01:01:01', 'bla' )")
    rel = duckdb_conn.table("test")
    rel.execute()
    try:
        rel.description
        queue.put(True)
    except:
        queue.put(False)


def cursor(duckdb_conn, queue, pandas):
    # Get a new connection
    cx = duckdb_conn.cursor()
    try:
        cx.execute('CREATE TABLE test (i bool, j TIME, k VARCHAR)')
        queue.put(False)
    except:
        queue.put(True)


class TestDuckMultithread(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_execute(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, execute_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_execute_many(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, execute_many_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_fetchone(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, fetchone_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_fetchall(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, fetchall_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_close(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, conn_close, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_fetchnp(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, fetchnp_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_fetchdf(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, fetchdf_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_fetchdfchunk(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, fetchdf_chunk_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_fetcharrow(self, duckdb_cursor, pandas):
        if not can_run:
            return
        duck_threads = DuckDBThreaded(10, fetch_arrow_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_fetch_record_batch(self, duckdb_cursor, pandas):
        if not can_run:
            return
        duck_threads = DuckDBThreaded(10, fetch_record_batch_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_transaction(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, transaction_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_df_append(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, df_append, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_df_register(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, df_register, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_df_unregister(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, df_unregister, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_arrow_register_unregister(self, duckdb_cursor, pandas):
        if not can_run:
            return
        duck_threads = DuckDBThreaded(10, arrow_register_unregister, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_table(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, table, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_view(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, view, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_values(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, values, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_from_query(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, from_query, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_from_DF(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, from_df, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_from_arrow(self, duckdb_cursor, pandas):
        if not can_run:
            return
        duck_threads = DuckDBThreaded(10, from_arrow, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_from_csv_auto(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, from_csv_auto, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_from_parquet(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, from_parquet, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_description(self, duckdb_cursor, pandas):
        duck_threads = DuckDBThreaded(10, description, pandas)
        duck_threads.multithread_test()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_cursor(self, duckdb_cursor, pandas):
        def only_some_succeed(results: List[bool]):
            if not any([result == True for result in results]):
                return False
            if all([result == True for result in results]):
                return False
            return True

        duck_threads = DuckDBThreaded(10, cursor, pandas)
        duck_threads.multithread_test(only_some_succeed)



import duckdb
import datetime
import numpy as np
import pytest
from conftest import NumpyPandas, ArrowPandas
from packaging.version import Version

pd = pytest.importorskip("pandas")


class TestDateTimeTimeStamp(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_timestamp_high(self, pandas, duckdb_cursor):
        duckdb_time = duckdb_cursor.sql("SELECT '2260-01-01 23:59:00'::TIMESTAMP AS '0'").df()
        df_in = pandas.DataFrame(
            {
                0: pandas.Series(
                    data=[datetime.datetime(year=2260, month=1, day=1, hour=23, minute=59)],
                    dtype='datetime64[us]',
                )
            }
        )
        df_out = duckdb_cursor.sql("select * from df_in").df()
        pandas.testing.assert_frame_equal(df_out, duckdb_time)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_timestamp_low(self, pandas, duckdb_cursor):
        duckdb_time = duckdb_cursor.sql(
            """
            SELECT '1680-01-01 23:59:00.234243'::TIMESTAMP AS '0'
        """
        ).df()
        df_in = pandas.DataFrame(
            {
                '0': pandas.Series(
                    data=[
                        pandas.Timestamp(
                            datetime.datetime(year=1680, month=1, day=1, hour=23, minute=59, microsecond=234243),
                            unit='us',
                        )
                    ],
                    dtype='datetime64[us]',
                )
            }
        )
        print('original:', duckdb_time['0'].dtype)
        print('df_in:', df_in['0'].dtype)
        df_out = duckdb_cursor.sql("select * from df_in").df()
        print('df_out:', df_out['0'].dtype)
        pandas.testing.assert_frame_equal(df_out, duckdb_time)

    @pytest.mark.skipif(
        Version(pd.__version__) < Version('2.0.2'), reason="pandas < 2.0.2 does not properly convert timezones"
    )
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_timestamp_timezone_regular(self, pandas, duckdb_cursor):
        duckdb_time = duckdb_cursor.sql(
            """
            SELECT timestamp '2022-01-01 12:00:00' AT TIME ZONE 'Pacific/Easter' as "0"
        """
        ).df()

        offset = datetime.timedelta(hours=-2)
        timezone = datetime.timezone(offset)
        df_in = pandas.DataFrame(
            {
                0: pandas.Series(
                    data=[datetime.datetime(year=2022, month=1, day=1, hour=15, tzinfo=timezone)], dtype='object'
                )
            }
        )
        df_out = duckdb_cursor.sql("select * from df_in").df()
        print(df_out)
        print(duckdb_time)
        pandas.testing.assert_frame_equal(df_out, duckdb_time)

    @pytest.mark.skipif(
        Version(pd.__version__) < Version('2.0.2'), reason="pandas < 2.0.2 does not properly convert timezones"
    )
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_timestamp_timezone_negative_extreme(self, pandas, duckdb_cursor):
        duckdb_time = duckdb_cursor.sql(
            """
            SELECT timestamp '2022-01-01 12:00:00' AT TIME ZONE 'Chile/EasterIsland' as "0"
        """
        ).df()

        offset = datetime.timedelta(hours=-19)
        timezone = datetime.timezone(offset)

        df_in = pandas.DataFrame(
            {
                0: pandas.Series(
                    data=[datetime.datetime(year=2021, month=12, day=31, hour=22, tzinfo=timezone)], dtype='object'
                )
            }
        )
        df_out = duckdb_cursor.sql("select * from df_in").df()
        pandas.testing.assert_frame_equal(df_out, duckdb_time)

    @pytest.mark.skipif(
        Version(pd.__version__) < Version('2.0.2'), reason="pandas < 2.0.2 does not properly convert timezones"
    )
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_timestamp_timezone_positive_extreme(self, pandas, duckdb_cursor):
        duckdb_time = duckdb_cursor.sql(
            """
            SELECT timestamp '2021-12-31 23:00:00' AT TIME ZONE 'kea_CV' as "0"
        """
        ).df()

        # 'kea_CV' is 20 hours ahead of UTC
        offset = datetime.timedelta(hours=20)
        timezone = datetime.timezone(offset)

        df_in = pandas.DataFrame(
            {
                0: pandas.Series(
                    data=[datetime.datetime(year=2022, month=1, day=1, hour=19, tzinfo=timezone)], dtype='object'
                )
            }
        )
        df_out = duckdb_cursor.sql("""select * from df_in""").df()
        pandas.testing.assert_frame_equal(df_out, duckdb_time)



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import Row
import textwrap
import duckdb


class TestSparkReadParquet(object):
    def test_read_parquet(self, duckdb_cursor, spark, tmp_path):
        file_path = tmp_path / 'basic.parquet'
        file_path = file_path.as_posix()
        duckdb_cursor.execute(
            f"COPY (select 42 a, true b, 'this is a long string' c) to '{file_path}' (FORMAT PARQUET)"
        )
        df = spark.read.parquet(file_path)
        res = df.collect()
        assert res == [Row(a=42, b=True, c='this is a long string')]



# timestamp ms precision

import numpy
from datetime import datetime


class TestNumpyTimestampMilliseconds(object):
    def test_numpy_timestamp(self, duckdb_cursor):
        res = duckdb_cursor.execute("SELECT TIMESTAMP '2019-11-26 21:11:42.501' as test_time").fetchnumpy()
        assert res['test_time'] == numpy.datetime64('2019-11-26 21:11:42.501')


class TestTimestampMilliseconds(object):
    def test_numpy_timestamp(self, duckdb_cursor):
        res = duckdb_cursor.execute("SELECT TIMESTAMP '2019-11-26 21:11:42.501' as test_time").fetchone()[0]
        assert res == datetime.strptime('2019-11-26 21:11:42.501', '%Y-%m-%d %H:%M:%S.%f')



import duckdb
import pytest
import sys
from typing import Union


def make_annotated_function(type: str):
    def test_base():
        return None

    import types

    test_function = types.FunctionType(
        test_base.__code__, test_base.__globals__, test_base.__name__, test_base.__defaults__, test_base.__closure__
    )
    # Add the 'type' string as return_annotation
    test_function.__annotations__ = {'return': type}
    return test_function


def python_version_lower_than_3_10():
    import sys

    if sys.version_info[0] < 3:
        return True
    if sys.version_info[1] < 10:
        return True
    return False


class TestStringAnnotation(object):
    @pytest.mark.skipif(
        python_version_lower_than_3_10(), reason="inspect.signature(eval_str=True) only supported since 3.10 and higher"
    )
    @pytest.mark.parametrize(
        ['input', 'expected'],
        [
            ('str', 'VARCHAR'),
            ('list[str]', 'VARCHAR[]'),
            ('dict[str, str]', 'MAP(VARCHAR, VARCHAR)'),
            ('dict[Union[str, bool], str]', 'MAP(UNION(u1 VARCHAR, u2 BOOLEAN), VARCHAR)'),
        ],
    )
    def test_string_annotations(self, duckdb_cursor, input, expected):
        from inspect import signature

        func = make_annotated_function(input)
        sig = signature(func)
        assert sig.return_annotation.__class__ == str

        duckdb_cursor.create_function("foo", func)
        rel = duckdb_cursor.sql("select foo()")
        assert rel.types == [expected]



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when, lit, array_contains
from duckdb.experimental.spark.sql.functions import sum, avg, max, min, mean, count


@pytest.fixture
def df(spark):
    simpleData = [
        ("James", "Sales", "NY", 90000, 34, 10000),
        ("Michael", "Sales", "NY", 86000, 56, 20000),
        ("Robert", "Sales", "CA", 81000, 30, 23000),
        ("Maria", "Finance", "CA", 90000, 24, 23000),
    ]

    columns = ["employee_name", "department", "state", "salary", "age", "bonus"]
    dataframe = spark.createDataFrame(data=simpleData, schema=columns)
    yield dataframe


@pytest.fixture
def df2(spark):
    simpleData2 = [
        ("James", "Sales", "NY", 90000, 34, 10000),
        ("Maria", "Finance", "CA", 90000, 24, 23000),
        ("Jen", "Finance", "NY", 79000, 53, 15000),
        ("Jeff", "Marketing", "CA", 80000, 25, 18000),
        ("Kumar", "Marketing", "NY", 91000, 50, 21000),
    ]
    columns2 = ["employee_name", "department", "state", "salary", "age", "bonus"]
    dataframe = spark.createDataFrame(data=simpleData2, schema=columns2)
    yield dataframe


class TestDataFrameUnion(object):
    def test_merge_with_union(self, df, df2):
        unionDF = df.union(df2)
        res = unionDF.collect()
        assert res == [
            Row(employee_name='James', department='Sales', state='NY', salary=90000, age=34, bonus=10000),
            Row(employee_name='Michael', department='Sales', state='NY', salary=86000, age=56, bonus=20000),
            Row(employee_name='Robert', department='Sales', state='CA', salary=81000, age=30, bonus=23000),
            Row(employee_name='Maria', department='Finance', state='CA', salary=90000, age=24, bonus=23000),
            Row(employee_name='James', department='Sales', state='NY', salary=90000, age=34, bonus=10000),
            Row(employee_name='Maria', department='Finance', state='CA', salary=90000, age=24, bonus=23000),
            Row(employee_name='Jen', department='Finance', state='NY', salary=79000, age=53, bonus=15000),
            Row(employee_name='Jeff', department='Marketing', state='CA', salary=80000, age=25, bonus=18000),
            Row(employee_name='Kumar', department='Marketing', state='NY', salary=91000, age=50, bonus=21000),
        ]
        unionDF = df.unionAll(df2)
        res2 = unionDF.collect()
        assert res == res2

    def test_merge_without_duplicates(self, df, df2):
        # 'sort' has been added to make the result deterministic
        disDF = df.union(df2).distinct().sort(col("employee_name"))
        res = disDF.collect()
        assert res == [
            Row(employee_name='James', department='Sales', state='NY', salary=90000, age=34, bonus=10000),
            Row(employee_name='Jeff', department='Marketing', state='CA', salary=80000, age=25, bonus=18000),
            Row(employee_name='Jen', department='Finance', state='NY', salary=79000, age=53, bonus=15000),
            Row(employee_name='Kumar', department='Marketing', state='NY', salary=91000, age=50, bonus=21000),
            Row(employee_name='Maria', department='Finance', state='CA', salary=90000, age=24, bonus=23000),
            Row(employee_name='Michael', department='Sales', state='NY', salary=86000, age=56, bonus=20000),
            Row(employee_name='Robert', department='Sales', state='CA', salary=81000, age=30, bonus=23000),
        ]



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")
pd = pytest.importorskip("pandas")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when
import duckdb
import re
from pandas.testing import assert_frame_equal


@pytest.fixture
def pandasDF(spark):
    data = [['Scott', 50], ['Jeff', 45], ['Thomas', 54], ['Ann', 34]]
    # Create the pandas DataFrame
    df = pd.DataFrame(data, columns=['Name', 'Age'])
    yield df


class TestPandasDataFrame(object):
    def test_pd_conversion_basic(self, spark, pandasDF):
        sparkDF = spark.createDataFrame(pandasDF)
        res = sparkDF.collect()
        sparkDF.show()
        expected = [
            Row(Name='Scott', Age=50),
            Row(Name='Jeff', Age=45),
            Row(Name='Thomas', Age=54),
            Row(Name='Ann', Age=34),
        ]
        assert res == expected

    def test_pd_conversion_schema(self, spark, pandasDF):
        mySchema = StructType([StructField("First Name", StringType(), True), StructField("Age", IntegerType(), True)])
        sparkDF = spark.createDataFrame(pandasDF, schema=mySchema)
        sparkDF.show()
        res = sparkDF.collect()
        expected = "[Row(First Name='Scott', Age=50), Row(First Name='Jeff', Age=45), Row(First Name='Thomas', Age=54), Row(First Name='Ann', Age=34)]"
        assert str(res) == expected

    def test_spark_to_pandas_dataframe(self, spark, pandasDF):
        sparkDF = spark.createDataFrame(pandasDF)
        res = sparkDF.toPandas()
        assert_frame_equal(res, pandasDF)



import duckdb
import pytest


# A closed connection should invalidate all relation's methods
class TestRAPICloseConnRel(object):
    def test_close_conn_rel(self, duckdb_cursor):
        con = duckdb.connect()
        con.execute("CREATE TABLE items(item VARCHAR, value DECIMAL(10,2), count INTEGER)")
        con.execute("INSERT INTO items VALUES ('jeans', 20.0, 1), ('hammer', 42.2, 2)")
        rel = con.table("items")
        con.close()

        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            print(rel)
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            len(rel)
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.aggregate("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.any_value("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.apply("", "")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.arg_max("", "")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.arg_min("", "")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.arrow()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.avg("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.bit_and("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.bit_or("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.bit_xor("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.bitstring_agg("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.bool_and("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.bool_or("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.count("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.create("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.create_view("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.cume_dist("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.dense_rank("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.describe()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.df()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.distinct()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.execute()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.favg("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.fetchall()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.fetchnumpy()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.fetchone()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.filter("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.first("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.first_value("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.fsum("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.geomean("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.histogram("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.insert("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.insert_into("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.lag("", "")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.last("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.last_value("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.lead("", "")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            print(rel.limit(1))
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.list("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.map(lambda df: df['col0'].add(42).to_frame())
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.max("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.mean("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.median("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.min("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.mode("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.n_tile("", 1)
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.nth_value("", "", 1)
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.order("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.percent_rank("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.product("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.project("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.quantile("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.quantile_cont("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.quantile_disc("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.query("", "")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.rank("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.rank_dense("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.row_number("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.std("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.stddev("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.stddev_pop("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.stddev_samp("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.string_agg("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.sum("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.to_arrow_table()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.to_df()
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.var("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.var_pop("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.var_samp("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.variance("")
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            rel.write_csv("")

        con = duckdb.connect()
        con.execute("CREATE TABLE items(item VARCHAR, value DECIMAL(10,2), count INTEGER)")
        con.execute("INSERT INTO items VALUES ('jeans', 20.0, 1), ('hammer', 42.2, 2)")
        valid_rel = con.table("items")

        # Test these bad boys when left relation is valid
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            valid_rel.union(rel)
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            valid_rel.except_(rel)
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            valid_rel.intersect(rel)
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            valid_rel.join(rel.set_alias('rel'), "rel.items = valid_rel.items")

    def test_del_conn(self, duckdb_cursor):
        con = duckdb.connect()
        con.execute("CREATE TABLE items(item VARCHAR, value DECIMAL(10,2), count INTEGER)")
        con.execute("INSERT INTO items VALUES ('jeans', 20.0, 1), ('hammer', 42.2, 2)")
        rel = con.table("items")
        del con
        with pytest.raises(duckdb.ConnectionException, match='Connection has already been closed'):
            print(rel)



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when, lit, array_contains
from duckdb.experimental.spark.sql.functions import sum, avg, max, min, mean, count


class TestDataFrameGroupBy(object):
    def test_group_by(self, spark):
        simpleData = [
            ("James", "Sales", "NY", 90000, 34, 10000),
            ("Michael", "Sales", "NY", 86000, 56, 20000),
            ("Robert", "Sales", "CA", 81000, 30, 23000),
            ("Maria", "Finance", "CA", 90000, 24, 23000),
            ("Raman", "Finance", "CA", 99000, 40, 24000),
            ("Scott", "Finance", "NY", 83000, 36, 19000),
            ("Jen", "Finance", "NY", 79000, 53, 15000),
            ("Jeff", "Marketing", "CA", 80000, 25, 18000),
            ("Kumar", "Marketing", "NY", 91000, 50, 21000),
        ]

        schema = ["employee_name", "department", "state", "salary", "age", "bonus"]
        df = spark.createDataFrame(data=simpleData, schema=schema)

        df2 = df.groupBy("department").sum("salary").sort("department")
        res = df2.collect()
        expected = "[Row(department='Finance', sum(salary)=351000), Row(department='Marketing', sum(salary)=171000), Row(department='Sales', sum(salary)=257000)]"
        assert str(res) == expected

        df2 = df.groupBy("department").count().sort("department")
        res = df2.collect()
        assert (
            str(res)
            == "[Row(department='Finance', count_star()=4), Row(department='Marketing', count_star()=2), Row(department='Sales', count_star()=3)]"
        )

        df2 = df.groupBy("department").min("salary").sort("department")
        res = df2.collect()
        assert (
            str(res)
            == "[Row(department='Finance', min(salary)=79000), Row(department='Marketing', min(salary)=80000), Row(department='Sales', min(salary)=81000)]"
        )

        df2 = df.groupBy("department").max("salary").sort("department")
        res = df2.collect()
        assert (
            str(res)
            == "[Row(department='Finance', max(salary)=99000), Row(department='Marketing', max(salary)=91000), Row(department='Sales', max(salary)=90000)]"
        )

        df2 = df.groupBy("department").avg("salary").sort("department")
        res = df2.collect()
        assert (
            str(res)
            == "[Row(department='Finance', avg(salary)=87750.0), Row(department='Marketing', avg(salary)=85500.0), Row(department='Sales', avg(salary)=85666.66666666667)]"
        )

        df2 = df.groupBy("department").mean("salary").sort("department")
        res = df2.collect()
        assert (
            str(res)
            == "[Row(department='Finance', mean(salary)=87750.0), Row(department='Marketing', mean(salary)=85500.0), Row(department='Sales', mean(salary)=85666.66666666667)]"
        )

        df2 = df.groupBy("department", "state").sum("salary", "bonus").sort("department", "state")
        res = df2.collect()
        assert (
            str(res)
            == "[Row(department='Finance', state='CA', sum(salary)=189000, sum(bonus)=47000), Row(department='Finance', state='NY', sum(salary)=162000, sum(bonus)=34000), Row(department='Marketing', state='CA', sum(salary)=80000, sum(bonus)=18000), Row(department='Marketing', state='NY', sum(salary)=91000, sum(bonus)=21000), Row(department='Sales', state='CA', sum(salary)=81000, sum(bonus)=23000), Row(department='Sales', state='NY', sum(salary)=176000, sum(bonus)=30000)]"
        )

        df2 = (
            df.groupBy("department")
            .agg(
                sum("salary").alias("sum_salary"),
                avg("salary").alias("avg_salary"),
                sum("bonus").alias("sum_bonus"),
                max("bonus").alias("max_bonus"),
            )
            .sort("department")
        )
        res = df2.collect()
        assert (
            str(res)
            == "[Row(department='Finance', sum_salary=351000, avg_salary=87750.0, sum_bonus=81000, max_bonus=24000), Row(department='Marketing', sum_salary=171000, avg_salary=85500.0, sum_bonus=39000, max_bonus=21000), Row(department='Sales', sum_salary=257000, avg_salary=85666.66666666667, sum_bonus=53000, max_bonus=23000)]"
        )

        df2 = (
            df.groupBy("department")
            .agg(
                sum("salary").alias("sum_salary"),
                avg("salary").alias("avg_salary"),
                sum("bonus").alias("sum_bonus"),
                max("bonus").alias("max_bonus"),
            )
            .where(col("sum_bonus") >= 50000)
            .sort("department")
        )
        res = df2.collect()
        print(str(res))
        assert (
            str(res)
            == "[Row(department='Finance', sum_salary=351000, avg_salary=87750.0, sum_bonus=81000, max_bonus=24000), Row(department='Sales', sum_salary=257000, avg_salary=85666.66666666667, sum_bonus=53000, max_bonus=23000)]"
        )



from .session import SparkSession
from .readwriter import DataFrameWriter
from .dataframe import DataFrame
from .conf import RuntimeConfig
from .catalog import Catalog

__all__ = ["SparkSession", "DataFrame", "RuntimeConfig", "DataFrameWriter", "Catalog"]



import numpy
import pandas


class TestHugeint(object):
    def test_hugeint(self, duckdb_cursor):
        duckdb_cursor.execute('SELECT 437894723897234238947043214')
        result = duckdb_cursor.fetchall()
        assert result == [(437894723897234238947043214,)]

    def test_hugeint_numpy(self, duckdb_cursor):
        duckdb_cursor.execute('SELECT 1::HUGEINT AS i')
        result = duckdb_cursor.fetchnumpy()
        assert result == {'i': numpy.array([1.0])}



import duckdb
import os
import pytest

pl = pytest.importorskip("polars")


def using_table(con, to_scan, object_name):
    exec(f"{object_name} = to_scan")
    return con.table(object_name)


def using_sql(con, to_scan, object_name):
    exec(f"{object_name} = to_scan")
    return con.sql(f"select * from to_scan")


# Fetch methods


def fetch_polars(rel):
    return rel.pl()


def fetch_df(rel):
    return rel.df()


def fetch_arrow(rel):
    return rel.arrow()


def fetch_arrow_table(rel):
    return rel.fetch_arrow_table()


def fetch_arrow_record_batch(rel):
    # Note: this has to executed first, otherwise we'll create a deadlock
    # Because it will try to execute the input at the same time as executing the relation
    # On the same connection (that's the core of the issue)
    return rel.execute().record_batch()


def fetch_relation(rel):
    return rel


class TestReplacementScan(object):
    def test_csv_replacement(self):
        con = duckdb.connect()
        filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'integers.csv')
        res = con.execute("select count(*) from '%s'" % (filename))
        assert res.fetchone()[0] == 2

    def test_parquet_replacement(self):
        con = duckdb.connect()
        filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'binary_string.parquet')
        res = con.execute("select count(*) from '%s'" % (filename))
        assert res.fetchone()[0] == 3

    @pytest.mark.parametrize('get_relation', [using_table, using_sql])
    @pytest.mark.parametrize(
        'fetch_method',
        [fetch_polars, fetch_df, fetch_arrow, fetch_arrow_table, fetch_arrow_record_batch, fetch_relation],
    )
    @pytest.mark.parametrize('object_name', ['tbl', 'table', 'select', 'update'])
    def test_table_replacement_scans(self, duckdb_cursor, get_relation, fetch_method, object_name):
        base_rel = duckdb_cursor.values([1, 2, 3])
        to_scan = fetch_method(base_rel)
        exec(f"{object_name} = to_scan")

        rel = get_relation(duckdb_cursor, to_scan, object_name)
        res = rel.fetchall()
        assert res == [(1, 2, 3)]

    def test_replacement_scan_relapi(self):
        con = duckdb.connect()
        pyrel1 = con.query('from (values (42), (84), (120)) t(i)')
        assert isinstance(pyrel1, duckdb.DuckDBPyRelation)
        assert pyrel1.fetchall() == [(42,), (84,), (120,)]

        pyrel2 = con.query('from pyrel1 limit 2')
        assert isinstance(pyrel2, duckdb.DuckDBPyRelation)
        assert pyrel2.fetchall() == [(42,), (84,)]

        pyrel3 = con.query('select i + 100 from pyrel2')
        assert type(pyrel3) == duckdb.DuckDBPyRelation
        assert pyrel3.fetchall() == [(142,), (184,)]

    def test_replacement_scan_alias(self):
        con = duckdb.connect()
        pyrel1 = con.query('from (values (1, 2)) t(i, j)')
        pyrel2 = con.query('from (values (1, 10)) t(i, k)')
        pyrel3 = con.query('from pyrel1 join pyrel2 using(i)')
        assert type(pyrel3) == duckdb.DuckDBPyRelation
        assert pyrel3.fetchall() == [(1, 2, 10)]

    def test_replacement_scan_pandas_alias(self):
        con = duckdb.connect()
        df1 = con.query('from (values (1, 2)) t(i, j)').df()
        df2 = con.query('from (values (1, 10)) t(i, k)').df()
        df3 = con.query('from df1 join df2 using(i)')
        assert df3.fetchall() == [(1, 2, 10)]

    def test_replacement_scan_fail(self):
        random_object = "I love salmiak rondos"
        con = duckdb.connect()
        with pytest.raises(
            duckdb.InvalidInputException,
            match=r'Python Object "random_object" of type "str" found on line .* not suitable for replacement scans.',
        ):
            con.execute("select count(*) from random_object").fetchone()



import duckdb
import pytest

pa = pytest.importorskip('pyarrow')


class TestArrowFetchRecordBatch(object):
    # Test with basic numeric conversion (integers, floats, and others fall this code-path)
    def test_record_batch_next_batch_numeric(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor_check = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select range a from range(3000);")
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)
        assert record_batch_reader.schema.names == ['a']
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 952
        with pytest.raises(StopIteration):
            chunk = record_batch_reader.read_next_batch()
        # Check if we are producing the correct thing
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)

        res = duckdb_cursor_check.execute("select * from record_batch_reader").fetchall()
        correct = duckdb_cursor.execute("select * from t").fetchall()
        assert res == correct

    # Test With Bool
    def test_record_batch_next_batch_bool(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor_check = duckdb.connect()
        duckdb_cursor.execute(
            "CREATE table t as SELECT CASE WHEN i % 2 = 0 THEN true ELSE false END AS a from range(3000) as tbl(i);"
        )
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)
        assert record_batch_reader.schema.names == ['a']
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 952
        with pytest.raises(StopIteration):
            chunk = record_batch_reader.read_next_batch()

        # Check if we are producing the correct thing
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)

        res = duckdb_cursor_check.execute("select * from record_batch_reader").fetchall()
        correct = duckdb_cursor.execute("select * from t").fetchall()
        assert res == correct

    # Test with Varchar
    def test_record_batch_next_batch_varchar(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor_check = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select range::varchar a from range(3000);")
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)
        assert record_batch_reader.schema.names == ['a']
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 952
        with pytest.raises(StopIteration):
            chunk = record_batch_reader.read_next_batch()

        # Check if we are producing the correct thing
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)

        res = duckdb_cursor_check.execute("select * from record_batch_reader").fetchall()
        correct = duckdb_cursor.execute("select * from t").fetchall()
        assert res == correct

    # Test with Struct
    def test_record_batch_next_batch_struct(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor_check = duckdb.connect()
        duckdb_cursor.execute(
            "CREATE table t as select {'x': i, 'y': i::varchar, 'z': i+1} as a from range(3000)  as tbl(i);"
        )
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)
        assert record_batch_reader.schema.names == ['a']
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 952
        with pytest.raises(StopIteration):
            chunk = record_batch_reader.read_next_batch()

        # Check if we are producing the correct thing
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)

        res = duckdb_cursor_check.execute("select * from record_batch_reader").fetchall()
        correct = duckdb_cursor.execute("select * from t").fetchall()
        assert res == correct

    # Test with List
    def test_record_batch_next_batch_list(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor_check = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select [i,i+1] as a from range(3000)  as tbl(i);")
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)
        assert record_batch_reader.schema.names == ['a']
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 952
        with pytest.raises(StopIteration):
            chunk = record_batch_reader.read_next_batch()

        # Check if we are producing the correct thing
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)

        res = duckdb_cursor_check.execute("select * from record_batch_reader").fetchall()
        correct = duckdb_cursor.execute("select * from t").fetchall()

        assert res == correct

    # Test with Map
    def test_record_batch_next_batch_map(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor_check = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select map([i], [i+1]) as a from range(3000)  as tbl(i);")
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)
        assert record_batch_reader.schema.names == ['a']
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 952
        with pytest.raises(StopIteration):
            chunk = record_batch_reader.read_next_batch()

        # Check if we are producing the correct thing
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)

        res = duckdb_cursor_check.execute("select * from record_batch_reader").fetchall()
        correct = duckdb_cursor.execute("select * from t").fetchall()

        assert res == correct

    # Test with Null Values
    def test_record_batch_next_batch_with_null(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor_check = duckdb.connect()
        duckdb_cursor.execute(
            "CREATE table t as SELECT CASE WHEN i % 2 = 0 THEN i ELSE NULL END AS a from range(3000) as tbl(i);"
        )
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)
        assert record_batch_reader.schema.names == ['a']
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1024
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 952
        with pytest.raises(StopIteration):
            chunk = record_batch_reader.read_next_batch()

        # Check if we are producing the correct thing
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)

        res = duckdb_cursor_check.execute("select * from record_batch_reader").fetchall()
        correct = duckdb_cursor.execute("select * from t").fetchall()

        assert res == correct

    def test_record_batch_read_default(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select range a from range(3000);")
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch()
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 3000

    def test_record_batch_next_batch_multiple_vectors_per_chunk(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select range a from range(5000);")
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(2048)
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 2048
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 2048
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 904
        with pytest.raises(StopIteration):
            chunk = record_batch_reader.read_next_batch()

        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1)
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 1

        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(2000)
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 2000

    def test_record_batch_next_batch_multiple_vectors_per_chunk_error(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select range a from range(5000);")
        query = duckdb_cursor.execute("SELECT a FROM t")
        with pytest.raises(RuntimeError, match='Approximate Batch Size of Record Batch MUST be higher than 0'):
            record_batch_reader = query.fetch_record_batch(0)
        with pytest.raises(TypeError, match='incompatible function arguments'):
            record_batch_reader = query.fetch_record_batch(-1)

    def test_record_batch_reader_from_relation(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select range a from range(3000);")
        relation = duckdb_cursor.table('t')
        record_batch_reader = relation.record_batch()
        chunk = record_batch_reader.read_next_batch()
        assert len(chunk) == 3000

    def test_record_coverage(self, duckdb_cursor):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select range a from range(2048);")
        query = duckdb_cursor.execute("SELECT a FROM t")
        record_batch_reader = query.fetch_record_batch(1024)

        chunk = record_batch_reader.read_all()
        assert len(chunk) == 2048

    def test_record_batch_query_error(self):
        duckdb_cursor = duckdb.connect()
        duckdb_cursor.execute("CREATE table t as select 'foo' as a;")
        with pytest.raises(duckdb.ConversionException, match='Conversion Error'):
            # 'execute' materializes the result, causing the error directly
            query = duckdb_cursor.execute("SELECT cast(a as double) FROM t")
            record_batch_reader = query.fetch_record_batch(1024)
            record_batch_reader.read_next_batch()

    def test_many_list_batches(self):
        conn = duckdb.connect()

        conn.execute(
            """
            create or replace table tbl as select * from (select {'a': [5,4,3,2,1]}), range(10000000)
        """
        )

        query = "SELECT * FROM tbl"
        chunk_size = 1_000_000

        # Because this produces multiple chunks, this caused a segfault before
        # because we changed some data in the first batch fetch
        batch_iter = conn.execute(query).fetch_record_batch(chunk_size)
        for batch in batch_iter:
            del batch

    def test_many_chunk_sizes(self):
        object_size = 1000000
        duckdb_cursor = duckdb.connect()
        query = duckdb_cursor.execute(f"CREATE table t as select range a from range({object_size});")
        for i in [1, 2, 4, 8, 16, 32, 33, 77, 999, 999999]:
            query = duckdb_cursor.execute("SELECT a FROM t")
            record_batch_reader = query.fetch_record_batch(i)
            num_loops = int(object_size / i)
            for j in range(num_loops):
                assert record_batch_reader.schema.names == ['a']
                chunk = record_batch_reader.read_next_batch()
                assert len(chunk) == i
            remainder = object_size % i
            if remainder > 0:
                chunk = record_batch_reader.read_next_batch()
                assert len(chunk) == remainder
            with pytest.raises(StopIteration):
                chunk = record_batch_reader.read_next_batch()



from re import S
import duckdb
import os
import pytest
import tempfile
from conftest import pandas_supports_arrow_backend

pa = pytest.importorskip("pyarrow")
pq = pytest.importorskip("pyarrow.parquet")
ds = pytest.importorskip("pyarrow.dataset")
np = pytest.importorskip("numpy")


class TestArrowLargeOffsets(object):
    @pytest.mark.skip(reason="CI does not have enough memory to validate this")
    def test_large_lists(self, duckdb_cursor):
        ary = pa.array([np.arange(start=0, stop=3000, dtype=np.uint8) for i in range(1_000_000)])
        tbl = pa.Table.from_pydict(dict(col=ary))
        with pytest.raises(
            duckdb.InvalidInputException,
            match='Arrow Appender: The maximum combined list offset for regular list buffers is 2147483647 but the offset of 2147481000 exceeds this.',
        ):
            res = duckdb_cursor.sql("SELECT col FROM tbl").arrow()

        tbl2 = pa.Table.from_pydict(dict(col=ary.cast(pa.large_list(pa.uint8()))))
        duckdb_cursor.sql("set arrow_large_buffer_size = true")
        res2 = duckdb_cursor.sql("SELECT col FROM tbl2").arrow()
        res2.validate()

    @pytest.mark.skip(reason="CI does not have enough memory to validate this")
    def test_large_maps(self, duckdb_cursor):
        ary = pa.array([np.arange(start=3000 * j, stop=3000 * (j + 1), dtype=np.uint64) for j in range(1_000_000)])
        tbl = pa.Table.from_pydict(dict(col=ary))

        with pytest.raises(
            duckdb.InvalidInputException,
            match='Arrow Appender: The maximum combined list offset for regular list buffers is 2147483647 but the offset of 2147481000 exceeds this.',
        ):
            arrow_map = duckdb_cursor.sql("select map(col, col) from tbl").arrow()

        duckdb_cursor.sql("set arrow_large_buffer_size = true")
        arrow_map_large = duckdb_cursor.sql("select map(col, col) from tbl").arrow()
        arrow_map_large.validate()



import pytest
import duckdb
import pandas as pd


class TestMultipleColumnsSameName(object):
    def test_multiple_columns_with_same_name(self, duckdb_cursor):
        df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [5, 6, 7, 8], 'd': [9, 10, 11, 12]})
        df = df.rename(columns={df.columns[1]: "a"})
        con = duckdb.connect()
        con.register('df_view', df)

        assert con.execute("DESCRIBE df_view;").fetchall() == [
            ('a', 'BIGINT', 'YES', None, None, None),
            ('a_1', 'BIGINT', 'YES', None, None, None),
            ('d', 'BIGINT', 'YES', None, None, None),
        ]
        assert con.execute("select a_1 from df_view;").fetchall() == [(5,), (6,), (7,), (8,)]
        assert con.execute("select a from df_view;").fetchall() == [(1,), (2,), (3,), (4,)]
        # Verify we are not changing original dataframe
        assert all(df.columns == ['a', 'a', 'd']), df.columns

    def test_multiple_columns_with_same_name_relation(self, duckdb_cursor):
        df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [5, 6, 7, 8], 'd': [9, 10, 11, 12]})
        df = df.rename(columns={df.columns[1]: "a"})
        con = duckdb.connect()
        rel = con.from_df(df)
        assert rel.query("df_view", "DESCRIBE df_view;").fetchall() == [
            ('a', 'BIGINT', 'YES', None, None, None),
            ('a_1', 'BIGINT', 'YES', None, None, None),
            ('d', 'BIGINT', 'YES', None, None, None),
        ]

        assert rel.query("df_view", "select a_1 from df_view;").fetchall() == [(5,), (6,), (7,), (8,)]
        assert rel.query("df_view", "select a from df_view;").fetchall() == [(1,), (2,), (3,), (4,)]

        # Verify we are not changing original dataframe
        assert all(df.columns == ['a', 'a', 'd']), df.columns

    def test_multiple_columns_with_same_name_replacement_scans(self, duckdb_cursor):
        df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [5, 6, 7, 8], 'd': [9, 10, 11, 12]})
        df = df.rename(columns={df.columns[1]: "a"})
        con = duckdb.connect()
        assert con.execute("select a_1 from df;").fetchall() == [(5,), (6,), (7,), (8,)]
        assert con.execute("select a from df;").fetchall() == [(1,), (2,), (3,), (4,)]
        # Verify we are not changing original dataframe
        assert all(df.columns == ['a', 'a', 'd']), df.columns

    def test_3669(self, duckdb_cursor):
        df = pd.DataFrame([(1, 5, 9), (2, 6, 10), (3, 7, 11), (4, 8, 12)], columns=['a_1', 'a', 'a'])
        con = duckdb.connect()
        con.register('df_view', df)
        assert con.execute("DESCRIBE df_view;").fetchall() == [
            ('a_1', 'BIGINT', 'YES', None, None, None),
            ('a', 'BIGINT', 'YES', None, None, None),
            ('a_2', 'BIGINT', 'YES', None, None, None),
        ]
        assert con.execute("select a_1 from df_view;").fetchall() == [(1,), (2,), (3,), (4,)]
        assert con.execute("select a from df_view;").fetchall() == [(5,), (6,), (7,), (8,)]
        # Verify we are not changing original dataframe
        assert all(df.columns == ['a_1', 'a', 'a']), df.columns

    def test_minimally_rename(self, duckdb_cursor):
        df = pd.DataFrame(
            [(1, 5, 9, 13), (2, 6, 10, 14), (3, 7, 11, 15), (4, 8, 12, 16)], columns=['a_1', 'a', 'a', 'a_2']
        )
        con = duckdb.connect()
        con.register('df_view', df)
        rel = con.sql("DESCRIBE df_view;")
        res = rel.fetchall()
        assert res == [
            ('a_1', 'BIGINT', 'YES', None, None, None),
            ('a', 'BIGINT', 'YES', None, None, None),
            ('a_2', 'BIGINT', 'YES', None, None, None),
            ('a_2_1', 'BIGINT', 'YES', None, None, None),
        ]
        assert con.execute("select a_1 from df_view;").fetchall() == [(1,), (2,), (3,), (4,)]
        assert con.execute("select a from df_view;").fetchall() == [(5,), (6,), (7,), (8,)]
        assert con.execute("select a_2 from df_view;").fetchall() == [(9,), (10,), (11,), (12,)]
        assert con.execute("select a_2_1 from df_view;").fetchall() == [(13,), (14,), (15,), (16,)]
        # Verify we are not changing original dataframe
        assert all(df.columns == ['a_1', 'a', 'a', 'a_2']), df.columns

    def test_multiple_columns_with_same_name_2(self, duckdb_cursor):
        df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [5, 6, 7, 8], 'a_1': [9, 10, 11, 12]})
        df = df.rename(columns={df.columns[1]: "a_1"})
        con = duckdb.connect()
        con.register('df_view', df)
        assert con.execute("DESCRIBE df_view;").fetchall() == [
            ('a', 'BIGINT', 'YES', None, None, None),
            ('a_1', 'BIGINT', 'YES', None, None, None),
            ('a_1_1', 'BIGINT', 'YES', None, None, None),
        ]
        assert con.execute("select a_1 from df_view;").fetchall() == [(5,), (6,), (7,), (8,)]
        assert con.execute("select a from df_view;").fetchall() == [(1,), (2,), (3,), (4,)]

        assert con.execute("select a_1_1 from df_view;").fetchall() == [(9,), (10,), (11,), (12,)]

    def test_case_insensitive(self, duckdb_cursor):
        df = pd.DataFrame({'A_1': [1, 2, 3, 4], 'a_1': [9, 10, 11, 12]})
        con = duckdb.connect()
        con.register('df_view', df)
        assert con.execute("DESCRIBE df_view;").fetchall() == [
            ('A_1', 'BIGINT', 'YES', None, None, None),
            ('a_1_1', 'BIGINT', 'YES', None, None, None),
        ]
        assert con.execute("select a_1 from df_view;").fetchall() == [(1,), (2,), (3,), (4,)]
        assert con.execute("select a_1_1 from df_view;").fetchall() == [(9,), (10,), (11,), (12,)]



# time type

import numpy
import datetime
import pandas


class TestNumpyTime(object):
    def test_fetchall_date(self, duckdb_cursor):
        res = duckdb_cursor.execute("SELECT TIME '13:06:40' as test_time").fetchall()
        assert res == [(datetime.time(13, 6, 40),)]

    def test_fetchnumpy_date(self, duckdb_cursor):
        res = duckdb_cursor.execute("SELECT TIME '13:06:40' as test_time").fetchnumpy()
        arr = numpy.array(['13:06:40'], dtype="object")
        arr = numpy.ma.masked_array(arr)
        numpy.testing.assert_array_equal(res['test_time'], arr)

    def test_fetchdf_date(self, duckdb_cursor):
        res = duckdb_cursor.execute("SELECT TIME '13:06:40' as test_time").fetchdf()
        ser = pandas.Series(numpy.array(['13:06:40'], dtype="object"), name="test_time")
        pandas.testing.assert_series_equal(res['test_time'], ser)



import duckdb
import pytest

pa = pytest.importorskip("pyarrow")
from duckdb.typing import *


class TestArrowBufferSize(object):
    def test_arrow_buffer_size(self):
        con = duckdb.connect()

        # All small string
        res = con.query("select 'bla'").arrow()
        assert res[0][0].type == pa.string()
        res = con.query("select 'bla'").record_batch()
        assert res.schema[0].type == pa.string()

        # All Large String
        con.execute("SET arrow_large_buffer_size=True")
        res = con.query("select 'bla'").arrow()
        assert res[0][0].type == pa.large_string()
        res = con.query("select 'bla'").record_batch()
        assert res.schema[0].type == pa.large_string()

        # All small string again
        con.execute("SET arrow_large_buffer_size=False")
        res = con.query("select 'bla'").arrow()
        assert res[0][0].type == pa.string()
        res = con.query("select 'bla'").record_batch()
        assert res.schema[0].type == pa.string()

    def test_arrow_buffer_size_udf(self):
        def just_return(x):
            return x

        con = duckdb.connect()
        con.create_function('just_return', just_return, [VARCHAR], VARCHAR, type='arrow')

        res = con.query("select just_return('bla')").arrow()

        assert res[0][0].type == pa.string()

        # All Large String
        con.execute("SET arrow_large_buffer_size=True")

        res = con.query("select just_return('bla')").arrow()
        assert res[0][0].type == pa.large_string()



import duckdb
import os
import datetime
import pytest

try:
    import pyarrow as pa
    import pandas as pd

    can_run = True
except:
    can_run = False


class TestArrowTimestamps(object):
    def test_timestamp_types(self, duckdb_cursor):
        if not can_run:
            return
        data = (
            pa.array([datetime.datetime.now()], type=pa.timestamp('ns')),
            pa.array([datetime.datetime.now()], type=pa.timestamp('us')),
            pa.array([datetime.datetime.now()], pa.timestamp('ms')),
            pa.array([datetime.datetime.now()], pa.timestamp('s')),
        )
        arrow_table = pa.Table.from_arrays([data[0], data[1], data[2], data[3]], ['a', 'b', 'c', 'd'])
        rel = duckdb.from_arrow(arrow_table).arrow()
        assert rel['a'] == arrow_table['a']
        assert rel['b'] == arrow_table['b']
        assert rel['c'] == arrow_table['c']
        assert rel['d'] == arrow_table['d']

    def test_timestamp_nulls(self, duckdb_cursor):
        if not can_run:
            return
        data = (
            pa.array([None], type=pa.timestamp('ns')),
            pa.array([None], type=pa.timestamp('us')),
            pa.array([None], pa.timestamp('ms')),
            pa.array([None], pa.timestamp('s')),
        )
        arrow_table = pa.Table.from_arrays([data[0], data[1], data[2], data[3]], ['a', 'b', 'c', 'd'])
        rel = duckdb.from_arrow(arrow_table).arrow()
        assert rel['a'] == arrow_table['a']
        assert rel['b'] == arrow_table['b']
        assert rel['c'] == arrow_table['c']
        assert rel['d'] == arrow_table['d']

    def test_timestamp_overflow(self, duckdb_cursor):
        if not can_run:
            return
        data = (
            pa.array([9223372036854775807], pa.timestamp('s')),
            pa.array([9223372036854775807], pa.timestamp('ms')),
            pa.array([9223372036854775807], pa.timestamp('us')),
        )
        arrow_table = pa.Table.from_arrays([data[0], data[1], data[2]], ['a', 'b', 'c'])
        arrow_from_duck = duckdb.from_arrow(arrow_table).arrow()
        assert arrow_from_duck['a'] == arrow_table['a']
        assert arrow_from_duck['b'] == arrow_table['b']
        assert arrow_from_duck['c'] == arrow_table['c']

        with pytest.raises(duckdb.ConversionException, match='Could not convert'):
            duck_rel = duckdb.from_arrow(arrow_table)
            res = duck_rel.project('a::TIMESTAMP_US')
            res.fetchone()

        with pytest.raises(duckdb.ConversionException, match='Could not convert'):
            duck_rel = duckdb.from_arrow(arrow_table)
            res = duck_rel.project('b::TIMESTAMP_US')
            res.fetchone()

        with pytest.raises(duckdb.ConversionException, match='Could not convert'):
            duck_rel = duckdb.from_arrow(arrow_table)
            res = duck_rel.project('c::TIMESTAMP_NS')
            res.fetchone()



""" The support for scaning over numpy arrays reuses many codes for pandas.
    Therefore, we only test the new codes and exec paths.
"""

import numpy as np
import duckdb
from datetime import timedelta
import pytest


class TestScanNumpy(object):
    def test_scan_numpy(self, duckdb_cursor):
        z = np.array([1, 2, 3])
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [(1,), (2,), (3,)]

        z = np.array([[1, 2, 3], [4, 5, 6]])
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [(1, 4), (2, 5), (3, 6)]

        z = [np.array([1, 2, 3]), np.array([4, 5, 6])]
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [(1, 4), (2, 5), (3, 6)]

        z = {"z": np.array([1, 2, 3]), "x": np.array([4, 5, 6])}
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [(1, 4), (2, 5), (3, 6)]

        z = np.array(["zzz", "xxx"])
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [('zzz',), ('xxx',)]

        z = [np.array(["zzz", "xxx"]), np.array([1, 2])]
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [('zzz', 1), ('xxx', 2)]

        # test ndarray with dtype = object (python dict)
        z = []
        for i in range(3):
            z.append({str(3 - i): i})
        z = np.array(z)
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [
            ({'key': ['3'], 'value': [0]},),
            ({'key': ['2'], 'value': [1]},),
            ({'key': ['1'], 'value': [2]},),
        ]

        # test timedelta
        delta = timedelta(days=50, seconds=27, microseconds=10, milliseconds=29000, minutes=5, hours=8, weeks=2)
        delta2 = timedelta(days=5, seconds=27, microseconds=10, milliseconds=29000, minutes=5, hours=8, weeks=2)
        z = np.array([delta, delta2])
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [
            (timedelta(days=64, seconds=29156, microseconds=10),),
            (timedelta(days=19, seconds=29156, microseconds=10),),
        ]

        # np.empty
        z = np.empty((3,))
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert len(res) == 3

        # empty list
        z = np.array([])
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == []

        # np.nan
        z = np.array([np.nan])
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [(None,)]

        # dict of mixed types
        z = {"z": np.array([1, 2, 3]), "x": np.array(["z", "x", "c"])}
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [(1, 'z'), (2, 'x'), (3, 'c')]

        # list of mixed types
        z = [np.array([1, 2, 3]), np.array(["z", "x", "c"])]
        res = duckdb_cursor.sql("select * from z").fetchall()
        assert res == [(1, 'z'), (2, 'x'), (3, 'c')]

        # currently unsupported formats, will throw duckdb.InvalidInputException

        # list of arrays with different length
        z = [np.array([1, 2]), np.array([3])]
        with pytest.raises(duckdb.InvalidInputException):
            duckdb_cursor.sql("select * from z")

        # dict of ndarrays of different length
        z = {"z": np.array([1, 2]), "x": np.array([3])}
        with pytest.raises(duckdb.InvalidInputException):
            duckdb_cursor.sql("select * from z")

        # high dimensional tensors
        z = np.array([[[1, 2]]])
        with pytest.raises(duckdb.InvalidInputException):
            duckdb_cursor.sql("select * from z")

        # list of ndarrys with len(shape) > 1
        z = [np.array([[1, 2], [3, 4]])]
        with pytest.raises(duckdb.InvalidInputException):
            duckdb_cursor.sql("select * from z")

        # dict of ndarrays with len(shape) > 1
        z = {"x": np.array([[1, 2], [3, 4]])}
        with pytest.raises(duckdb.InvalidInputException):
            duckdb_cursor.sql("select * from z")



import json
import os
import sys
import webbrowser
from functools import reduce
import argparse

qgraph_css = """
.styled-table {
	border-collapse: collapse;
	margin: 25px 0;
	font-size: 0.9em;
	font-family: sans-serif;
	min-width: 400px;
	box-shadow: 0 0 20px rgba(0, 0, 0, 0.15);
}
.styled-table thead tr {
	background-color: #009879;
	color: #ffffff;
	text-align: left;
}
.styled-table th,
.styled-table td {
	padding: 12px 15px;
}
.styled-table tbody tr {
	border-bottom: 1px solid #dddddd;
}

.styled-table tbody tr:nth-of-type(even) {
	background-color: #f3f3f3;
}

.styled-table tbody tr:last-of-type {
	border-bottom: 2px solid #009879;
}

.node-body {
	font-size:15px;
}
.tf-nc {
	position: relative;
	width: 250px;
	text-align: center;
	background-color: #fff100;
}
"""


class NodeTiming:

    def __init__(self, phase: str, time: float) -> object:
        self.phase = phase
        self.time = time
        # percentage is determined later.
        self.percentage = 0

    def calculate_percentage(self, total_time: float) -> None:
        self.percentage = self.time / total_time

    def combine_timing(l: object, r: object) -> object:
        # TODO: can only add timings for same-phase nodes
        total_time = l.time + r.time
        return NodeTiming(l.phase, total_time)


class AllTimings:

    def __init__(self):
        self.phase_to_timings = {}

    def add_node_timing(self, node_timing: NodeTiming):
        if node_timing.phase in self.phase_to_timings:
            self.phase_to_timings[node_timing.phase].append(node_timing)
            return
        self.phase_to_timings[node_timing.phase] = [node_timing]

    def get_phase_timings(self, phase: str):
        return self.phase_to_timings[phase]

    def get_summary_phase_timings(self, phase: str):
        return reduce(NodeTiming.combine_timing, self.phase_to_timings[phase])

    def get_phases(self):
        phases = list(self.phase_to_timings.keys())
        phases.sort(key=lambda x: (self.get_summary_phase_timings(x)).time)
        phases.reverse()
        return phases

    def get_sum_of_all_timings(self):
        total_timing_sum = 0
        for phase in self.phase_to_timings.keys():
            total_timing_sum += self.get_summary_phase_timings(phase).time
        return total_timing_sum


def open_utf8(fpath: str, flags: str) -> object:
    return open(fpath, flags, encoding="utf8")


def get_child_timings(top_node: object, query_timings: object) -> str:
    node_timing = NodeTiming(top_node['name'], float(top_node['timing']))
    query_timings.add_node_timing(node_timing)
    for child in top_node['children']:
        get_child_timings(child, query_timings)


color_map = {
    "HASH_JOIN": "#ffffba",
    "PROJECTION": "#ffb3ba",
    "SEQ_SCAN": "#baffc9",
    "UNGROUPED_AGGREGATE": "#ffdfba",
    "FILTER": "#bae1ff",
    "ORDER_BY": "#facd60",
    "PERFECT_HASH_GROUP_BY": "#ffffba",
    "HASH_GROUP_BY": "#ffffba",
    "NESTED_LOOP_JOIN": "#ffffba",
    "STREAMING_LIMIT": "#facd60",
    "COLUMN_DATA_SCAN": "#1ac0c6",
    "TOP_N": "#ffdfba"
}


def get_node_body(name: str, result: str, cardinality: float, extra_info: str, timing: object) -> str:
    node_style = ""
    stripped_name = name.strip()
    if stripped_name in color_map:
        node_style = f"background-color: {color_map[stripped_name]};"

    body = f"<span class=\"tf-nc\" style=\"{node_style}\">"
    body += "<div class=\"node-body\">"
    new_name = name.replace("_", " ")
    body += f"<p> <b>{new_name} ({result}s) </b></p>"
    if extra_info:
        extra_info = extra_info.replace("[INFOSEPARATOR]", "----")
        extra_info = extra_info.replace("<br><br>", "<br>")
        body += f"<p> {extra_info} </p>"
    body += f"<p> cardinality = {cardinality} </p>"
    # TODO: Expand on timing. Usually available from a detailed profiling
    body += "</div>"
    body += "</span>"
    return body


def generate_tree_recursive(json_graph: object) -> str:
    node_prefix_html = "<li>"
    node_suffix_html = "</li>"
    node_body = get_node_body(json_graph["name"],
                              json_graph["timing"],
                              json_graph["cardinality"],
                              json_graph["extra_info"].replace("\n", "<br>"),
                              json_graph["timings"])

    children_html = ""
    if len(json_graph['children']) >= 1:
        children_html += "<ul>"
        for child in json_graph["children"]:
            children_html += generate_tree_recursive(child)
        children_html += "</ul>"
    return node_prefix_html + node_body + children_html + node_suffix_html


# For generating the table in the top left.
def generate_timing_html(graph_json: object, query_timings: object) -> object:
    json_graph = json.loads(graph_json)
    gather_timing_information(json_graph, query_timings)
    total_time = float(json_graph['timing'])
    table_head = """
	<table class=\"styled-table\"> 
		<thead>
			<tr>
				<th>Phase</th>
				<th>Time</th>
				<th>Percentage</th>
			</tr>
		</thead>"""

    table_body = "<tbody>"
    table_end = "</tbody></table>"

    execution_time = query_timings.get_sum_of_all_timings()

    all_phases = query_timings.get_phases()
    query_timings.add_node_timing(NodeTiming("TOTAL TIME", total_time))
    query_timings.add_node_timing(NodeTiming("Execution Time", execution_time))
    all_phases = ["TOTAL TIME", "Execution Time"] + all_phases
    for phase in all_phases:
        summarized_phase = query_timings.get_summary_phase_timings(phase)
        summarized_phase.calculate_percentage(total_time)
        phase_column = f"<b>{phase}</b>" if phase == "TOTAL TIME" or phase == "Execution Time" else phase
        table_body += f"""
	<tr>
			<td>{phase_column}</td>
            <td>{summarized_phase.time}</td>
            <td>{str(summarized_phase.percentage * 100)[:6]}%</td>
    </tr>
"""
    table_body += table_end
    return table_head + table_body


def generate_tree_html(graph_json: object) -> str:
    json_graph = json.loads(graph_json)
    tree_prefix = "<div class=\"tf-tree tf-gap-lg\"> \n <ul>"
    tree_suffix = "</ul> </div>"
    # first level of json is general overview
    # FIXME: make sure json output first level always has only 1 level
    tree_body = generate_tree_recursive(json_graph['children'][0])
    return tree_prefix + tree_body + tree_suffix


def generate_ipython(json_input: str) -> str:
    from IPython.core.display import HTML

    html_output = generate_html(json_input, False)

    return HTML(("\n"
                 "	${CSS}\n"
                 "	${LIBRARIES}\n"
                 "	<div class=\"chart\" id=\"query-profile\"></div>\n"
                 "	${CHART_SCRIPT}\n"
                 "	").replace("${CSS}", html_output['css']).replace('${CHART_SCRIPT}',
                                                                       html_output['chart_script']).replace(
        '${LIBRARIES}', html_output['libraries']))


def generate_style_html(graph_json: str, include_meta_info: bool) -> None:
    treeflex_css = "<link rel=\"stylesheet\" href=\"https://unpkg.com/treeflex/dist/css/treeflex.css\">\n"
    css = "<style>\n"
    css += qgraph_css + "\n"
    css += "</style>\n"
    return {
        'treeflex_css': treeflex_css,
        'duckdb_css': css,
        'libraries': '',
        'chart_script': ''
    }


def gather_timing_information(json: str, query_timings: object) -> None:
    # add up all of the times
    # measure each time as a percentage of the total time.
    # then you can return a list of [phase, time, percentage]
    get_child_timings(json['children'][0], query_timings)


def translate_json_to_html(input_file: str, output_file: str) -> None:
    query_timings = AllTimings()
    with open_utf8(input_file, 'r') as f:
        text = f.read()

    html_output = generate_style_html(text, True)
    timing_table = generate_timing_html(text, query_timings)
    tree_output = generate_tree_html(text)

    # finally create and write the html
    with open_utf8(output_file, "w+") as f:
        html = """<!DOCTYPE html>
<html>
	<head>
	<meta charset="utf-8">
	<meta name="viewport" content="width=device-width">
	<title>Query Profile Graph for Query</title>
	${TREEFLEX_CSS}
	<style>
		${DUCKDB_CSS}
	</style>
</head>
<body>
	<div id="meta-info"></div>
	<div class="chart" id="query-profile">
		${TIMING_TABLE}
	</div>
	${TREE}
</body>
</html>
"""
        html = html.replace("${TREEFLEX_CSS}", html_output['treeflex_css'])
        html = html.replace("${DUCKDB_CSS}", html_output['duckdb_css'])
        html = html.replace("${TIMING_TABLE}", timing_table)
        html = html.replace('${TREE}', tree_output)
        f.write(html)


def main() -> None:
    if sys.version_info[0] < 3:
        print("Please use python3")
        exit(1)
    parser = argparse.ArgumentParser(
        prog='Query Graph Generator',
        description='Given a json profile output, generate a html file showing the query graph and timings of operators')
    parser.add_argument('profile_input', help='profile input in json')
    parser.add_argument('--out', required=False, default=False)
    parser.add_argument('--open', required=False, action='store_true', default=True)
    args = parser.parse_args()

    input = args.profile_input
    output = args.out
    if not args.out:
        if ".json" in input:
            output = input.replace(".json", ".html")
        else:
            print("please provide profile output in json")
            exit(1)
    else:
        if ".html" in args.out:
            output = args.out
        else:
            print("please provide valid .html file for output name")
            exit(1)

    open_output = args.open

    translate_json_to_html(input, output)

    if open_output:
        webbrowser.open('file://' + os.path.abspath(output), new=2)


if __name__ == '__main__':
    main()



import duckdb
import os
import shutil


class TestMultiStatement(object):
    def test_multi_statement(self, duckdb_cursor):
        import duckdb

        con = duckdb.connect(':memory:')

        # test empty statement
        con.execute('')

        # run multiple statements in one call to execute
        con.execute(
            '''
        CREATE TABLE integers(i integer);
        insert into integers select * from range(10);
        select * from integers;
        '''
        )
        results = [x[0] for x in con.fetchall()]
        assert results == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

        # test export/import
        export_location = os.path.join(os.getcwd(), 'duckdb_pytest_dir_export')
        try:
            shutil.rmtree(export_location)
        except:
            pass
        con.execute('CREATE TABLE integers2(i INTEGER)')
        con.execute('INSERT INTO integers2 VALUES (1), (5), (7), (1928)')
        con.execute("EXPORT DATABASE '%s'" % (export_location,))
        # reset connection
        con = duckdb.connect(':memory:')
        con.execute("IMPORT DATABASE '%s'" % (export_location,))
        integers = [x[0] for x in con.execute('SELECT * FROM integers').fetchall()]
        integers2 = [x[0] for x in con.execute('SELECT * FROM integers2').fetchall()]
        assert integers == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        assert integers2 == [1, 5, 7, 1928]
        shutil.rmtree(export_location)



import duckdb
import os
import pytest

from conftest import pandas_supports_arrow_backend

pa = pytest.importorskip("pyarrow")
ds = pytest.importorskip("pyarrow.dataset")
_ = pytest.importorskip("pandas", '2.0.0')


@pytest.mark.skipif(not pandas_supports_arrow_backend(), reason="pandas does not support the 'pyarrow' backend")
class TestArrowDFProjectionPushdown(object):
    def test_projection_pushdown_no_filter(self, duckdb_cursor):
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("CREATE TABLE test (a  INTEGER, b INTEGER, c INTEGER)")
        duckdb_conn.execute("INSERT INTO  test VALUES (1,1,1),(10,10,10),(100,10,100),(NULL,NULL,NULL)")
        duck_tbl = duckdb_conn.table("test")
        arrow_table = duck_tbl.df().convert_dtypes(dtype_backend='pyarrow')
        duckdb_conn.register("testarrowtable", arrow_table)
        assert duckdb_conn.execute("SELECT sum(a) FROM  testarrowtable").fetchall() == [(111,)]



import pandas as pd
import duckdb
import numpy as np


class TestPandasStride(object):
    def test_stride(self, duckdb_cursor):
        expected_df = pd.DataFrame(np.arange(20).reshape(5, 4), columns=["a", "b", "c", "d"])
        con = duckdb.connect()
        con.register('df_view', expected_df)
        output_df = con.execute("SELECT * FROM df_view;").fetchdf()
        pd.testing.assert_frame_equal(expected_df, output_df)

    def test_stride_fp32(self, duckdb_cursor):
        expected_df = pd.DataFrame(np.arange(20, dtype='float32').reshape(5, 4), columns=["a", "b", "c", "d"])
        con = duckdb.connect()
        con.register('df_view', expected_df)
        output_df = con.execute("SELECT * FROM df_view;").fetchdf()
        for col in output_df.columns:
            assert str(output_df[col].dtype) == 'float32'
        pd.testing.assert_frame_equal(expected_df, output_df)

    def test_stride_fp64(self, duckdb_cursor):
        expected_df = pd.DataFrame(np.arange(20, dtype='float64').reshape(5, 4), columns=["a", "b", "c", "d"])
        con = duckdb.connect()
        con.register('df_view', expected_df)
        output_df = con.execute("SELECT * FROM df_view;").fetchdf()
        for col in output_df.columns:
            assert str(output_df[col].dtype) == 'float64'
        pd.testing.assert_frame_equal(expected_df, output_df)



import duckdb
import os
import pytest

pyarrow_parquet = pytest.importorskip("pyarrow.parquet")
import sys


class TestProgressBarArrow(object):
    def test_progress_arrow(self):
        if os.name == 'nt':
            return
        np = pytest.importorskip("numpy")
        pyarrow = pytest.importorskip("pyarrow")

        data = pyarrow.array(np.arange(10000000), type=pyarrow.int32())
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA progress_bar_time=1")
        duckdb_conn.execute("PRAGMA disable_print_progress_bar")

        tbl = pyarrow.Table.from_arrays([data], ['a'])
        rel = duckdb_conn.from_arrow(tbl)
        result = rel.aggregate('sum(a)')
        assert result.execute().fetchone()[0] == 49999995000000
        # Multiple Threads
        duckdb_conn.execute("PRAGMA threads=4")
        duckdb_conn.execute("PRAGMA verify_parallelism")
        assert result.execute().fetchone()[0] == 49999995000000

        # More than one batch
        tbl = pyarrow.Table.from_batches(pyarrow.Table.from_arrays([data], ['a']).to_batches(100))
        rel = duckdb_conn.from_arrow(tbl)
        result = rel.aggregate('sum(a)')
        assert result.execute().fetchone()[0] == 49999995000000

        # Single Thread
        duckdb_conn.execute("PRAGMA threads=1")
        duck_res = result.execute()
        py_res = duck_res.fetchone()[0]
        assert py_res == 49999995000000

    def test_progress_arrow_empty(self):
        if os.name == 'nt':
            return
        np = pytest.importorskip("numpy")
        pyarrow = pytest.importorskip("pyarrow")

        data = pyarrow.array(np.arange(0), type=pyarrow.int32())
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA progress_bar_time=1")
        duckdb_conn.execute("PRAGMA disable_print_progress_bar")

        tbl = pyarrow.Table.from_arrays([data], ['a'])
        rel = duckdb_conn.from_arrow(tbl)
        result = rel.aggregate('sum(a)')
        assert result.execute().fetchone()[0] == None



# simple DB API testcase

import pytest
import duckdb


class TestDBAPICursor(object):
    def test_cursor_basic(self):
        # Create a connection
        con = duckdb.connect(':memory:')
        # Then create a cursor on the connection
        cursor = con.cursor()
        # Use the cursor for queries
        res = cursor.execute("select [1,2,3,NULL,4]").fetchall()
        assert res == [([1, 2, 3, None, 4],)]

    def test_cursor_preexisting(self):
        con = duckdb.connect(':memory:')
        con.execute("create table tbl as select i a, i+1 b, i+2 c from range(5) tbl(i)")
        cursor = con.cursor()
        res = cursor.execute("select * from tbl").fetchall()
        assert res == [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, 5), (4, 5, 6)]

    def test_cursor_after_creation(self):
        con = duckdb.connect(':memory:')
        # First create the cursor
        cursor = con.cursor()
        # Then create table on the source connection
        con.execute("create table tbl as select i a, i+1 b, i+2 c from range(5) tbl(i)")
        res = cursor.execute("select * from tbl").fetchall()
        assert res == [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, 5), (4, 5, 6)]

    def test_cursor_mixed(self):
        con = duckdb.connect(':memory:')
        # First create the cursor
        cursor = con.cursor()
        # Then create table on the cursor
        cursor.execute("create table tbl as select i a, i+1 b, i+2 c from range(5) tbl(i)")
        # Close the cursor and create a new one
        cursor.close()
        cursor = con.cursor()
        res = cursor.execute("select * from tbl").fetchall()
        assert res == [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, 5), (4, 5, 6)]

    def test_cursor_temp_schema_closed(self):
        con = duckdb.connect(':memory:')
        cursor = con.cursor()
        cursor.execute("create temp table tbl as select * from range(100)")
        other_cursor = con.cursor()
        # Connection that created the table is closed
        cursor.close()
        with pytest.raises(duckdb.CatalogException):
            # This table does not exist in this cursor
            res = other_cursor.execute("select * from tbl").fetchall()

    def test_cursor_temp_schema_open(self):
        con = duckdb.connect(':memory:')
        cursor = con.cursor()
        cursor.execute("create temp table tbl as select * from range(100)")
        other_cursor = con.cursor()
        # Connection that created the table is still open
        # cursor.close()
        with pytest.raises(duckdb.CatalogException):
            # This table does not exist in this cursor
            res = other_cursor.execute("select * from tbl").fetchall()

    def test_cursor_temp_schema_both(self):
        con = duckdb.connect(':memory:')
        cursor1 = con.cursor()
        cursor2 = con.cursor()
        cursor3 = con.cursor()
        cursor1.execute("create temp table tbl as select i from range(10) tbl(i)")
        cursor2.execute("create temp table tbl as select i+10 from range(10) tbl(i)")
        with pytest.raises(duckdb.CatalogException):
            # This table does not exist in this cursor
            res = cursor3.execute("select * from tbl").fetchall()
        res = cursor1.execute("select * from tbl").fetchall()
        assert res == [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,)]
        res = cursor2.execute("select * from tbl").fetchall()
        assert res == [(10,), (11,), (12,), (13,), (14,), (15,), (16,), (17,), (18,), (19,)]

        cursor1.close()
        cursor2.close()

    def test_cursor_closed(self):
        con = duckdb.connect(':memory:')
        con.close()
        with pytest.raises(duckdb.ConnectionException):
            cursor = con.cursor()

    def test_cursor_used_after_connection_closed(self):
        con = duckdb.connect(':memory:')
        cursor = con.cursor()
        con.close()
        with pytest.raises(duckdb.ConnectionException):
            cursor.execute("select [1,2,3,4]")

    def test_cursor_used_after_close(self):
        con = duckdb.connect(':memory:')
        cursor = con.cursor()
        cursor.close()
        with pytest.raises(duckdb.ConnectionException):
            cursor.execute("select [1,2,3,4]")



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")


class TestSparkRuntimeConfig(object):
    def test_spark_runtime_config(self, spark):
        # This fetches the internal runtime config from the session
        spark.conf

    def test_spark_runtime_config_set(self, spark):
        # Set Config
        with pytest.raises(NotImplementedError):
            spark.conf.set("spark.executor.memory", "5g")

    @pytest.mark.skip(reason="RuntimeConfig is not implemented yet")
    def test_spark_runtime_config_get(self, spark):
        # Get a Spark Config
        with pytest.raises(KeyError):
            partitions = spark.conf.get("spark.sql.shuffle.partitions")



import duckdb
import numpy
import pytest
from conftest import NumpyPandas, ArrowPandas


def check_result_list(res):
    for res_item in res:
        assert res_item[0] == res_item[1]


def check_create_table(category, pandas):
    conn = duckdb.connect()

    conn.execute("PRAGMA enable_verification")
    df_in = pandas.DataFrame(
        {
            'x': pandas.Categorical(category, ordered=True),
            'y': pandas.Categorical(category, ordered=True),
            'z': category,
        }
    )

    category.append('bla')

    df_in_diff = pandas.DataFrame(
        {
            'k': pandas.Categorical(category, ordered=True),
        }
    )

    df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data")
    df_out = df_out.df()
    assert df_in.equals(df_out)

    conn.execute("CREATE TABLE t1 AS SELECT * FROM df_in")
    conn.execute("CREATE TABLE t2 AS SELECT * FROM df_in")

    # Check fetchall
    res = conn.execute("SELECT x,z FROM t1").fetchall()
    check_result_list(res)

    # Do a insert to trigger string -> cat
    conn.execute("INSERT INTO t1 VALUES ('2','2','2')")

    res = conn.execute("SELECT x FROM t1 where x = '1'").fetchall()
    assert res == [('1',)]

    res = conn.execute("SELECT t1.x FROM t1 inner join t2 on (t1.x = t2.x) order by t1.x").fetchall()
    assert res == conn.execute("SELECT x FROM t1 order by t1.x").fetchall()

    res = conn.execute("SELECT t1.x FROM t1 inner join t2 on (t1.x = t2.y) order by t1.x").fetchall()
    correct_res = conn.execute("SELECT x FROM t1 order by x").fetchall()
    assert res == correct_res

    # Run equal ENUM comparison
    res = conn.execute("SELECT t1.x FROM t1,t2 where t1.x = t2.x order by t1.x").fetchall()
    assert res == correct_res

    # Run different ENUM comparison
    conn.execute("CREATE TABLE t3 AS SELECT * FROM df_in_diff")
    res = conn.execute("SELECT t1.x FROM t1,t3 where t3.k = t1.x order by t1.x").fetchall()
    assert res == correct_res

    # Triggering the cast with ENUM as a src
    conn.execute("ALTER TABLE t1 ALTER x SET DATA TYPE VARCHAR")
    # We should be able to drop the table without any dependencies
    conn.execute("DROP TABLE t1")


# TODO: extend tests with ArrowPandas
class TestCategory(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_category_string_uint16(self, duckdb_cursor, pandas):
        category = []
        for i in range(300):
            category.append(str(i))
        check_create_table(category, pandas)

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_category_string_uint32(self, duckdb_cursor, pandas):
        category = []
        for i in range(70000):
            category.append(str(i))
        check_create_table(category, pandas)



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")
from duckdb.experimental.spark.sql import SparkSession


class TestSparkSession(object):
    def test_spark_session_default(self):
        session = SparkSession.builder.getOrCreate()

    def test_spark_session(self):
        session = SparkSession.builder.master("local[1]").appName('SparkByExamples.com').getOrCreate()

    def test_new_session(self, spark: SparkSession):
        session = spark.newSession()

    @pytest.mark.skip(reason='not tested yet')
    def test_retrieve_same_session(self):
        spark = SparkSession.builder.master('test').appName('test2').getOrCreate()
        spark2 = SparkSession.builder.getOrCreate()
        # Same connection should be returned
        assert spark == spark2

    def test_config(self):
        # Usage of config()
        spark = (
            SparkSession.builder.master("local[1]")
            .appName("SparkByExamples.com")
            .config("spark.some.config.option", "config-value")
            .getOrCreate()
        )

    @pytest.mark.skip(reason="enableHiveSupport is not implemented yet")
    def test_hive_support(self):
        # Enabling Hive to use in Spark
        spark = (
            SparkSession.builder.master("local[1]")
            .appName("SparkByExamples.com")
            .config("spark.sql.warehouse.dir", "<path>/spark-warehouse")
            .enableHiveSupport()
            .getOrCreate()
        )

    def test_version(self, spark):
        version = spark.version
        assert version == '1.0.0'

    def test_get_active_session(self, spark):
        active_session = spark.getActiveSession()

    def test_read(self, spark):
        reader = spark.read

    def test_write(self, spark):
        df = spark.sql('select 42')
        writer = df.write

    def test_read_stream(self, spark):
        reader = spark.readStream

    def test_spark_context(self, spark):
        context = spark.sparkContext

    def test_sql(self, spark):
        df = spark.sql('select 42')

    def test_stop_context(self, spark):
        context = spark.sparkContext
        spark.stop()

    def test_table(self, spark):
        spark.sql('create table tbl(a varchar(10))')
        df = spark.table('tbl')

    def test_udf(self, spark):
        with pytest.raises(NotImplementedError):
            udf_registration = spark.udf



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import duckdb
import pandas as pd


class TestUnicode(object):
    def test_unicode_pandas_scan(self, duckdb_cursor):
        con = duckdb.connect(database=':memory:', read_only=False)
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3], "j": ["a", "c", u"ë"]})
        con.register('test_df_view', test_df)
        con.execute('SELECT i, j, LENGTH(j) FROM test_df_view').fetchall()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import duckdb
import numpy
import datetime
import pytest
from conftest import NumpyPandas, ArrowPandas


def run_parallel_queries(main_table, left_join_table, expected_df, pandas, iteration_count=5):
    for i in range(0, iteration_count):
        output_df = None
        sql = """
        select
            main_table.*
            ,t1.*
            ,t2.*
        from main_table
        left join left_join_table t1
            on main_table.join_column = t1.join_column
        left join left_join_table t2
            on main_table.join_column = t2.join_column
        """
        try:
            duckdb_conn = duckdb.connect()
            duckdb_conn.execute("PRAGMA threads=4")
            duckdb_conn.register('main_table', main_table)
            duckdb_conn.register('left_join_table', left_join_table)
            output_df = duckdb_conn.execute(sql).fetchdf()
            pandas.testing.assert_frame_equal(expected_df, output_df)
            print(output_df)
        except Exception as err:
            print(err)
        finally:
            duckdb_conn.close()


class TestParallelPandasScan(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_parallel_numeric_scan(self, duckdb_cursor, pandas):
        main_table = pandas.DataFrame([{"join_column": 3}])
        left_join_table = pandas.DataFrame([{"join_column": 3, "other_column": 4}])
        run_parallel_queries(main_table, left_join_table, left_join_table, pandas)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_parallel_ascii_text(self, duckdb_cursor, pandas):
        main_table = pandas.DataFrame([{"join_column": "text"}])
        left_join_table = pandas.DataFrame([{"join_column": "text", "other_column": "more text"}])
        run_parallel_queries(main_table, left_join_table, left_join_table, pandas)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_parallel_unicode_text(self, duckdb_cursor, pandas):
        main_table = pandas.DataFrame([{"join_column": u"mühleisen"}])
        left_join_table = pandas.DataFrame([{"join_column": u"mühleisen", "other_column": u"höhöhö"}])
        run_parallel_queries(main_table, left_join_table, left_join_table, pandas)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_parallel_complex_unicode_text(self, duckdb_cursor, pandas):
        main_table = pandas.DataFrame([{"join_column": u"鴨"}])
        left_join_table = pandas.DataFrame([{"join_column": u"鴨", "other_column": u"數據庫"}])
        run_parallel_queries(main_table, left_join_table, left_join_table, pandas)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_parallel_emojis(self, duckdb_cursor, pandas):
        main_table = pandas.DataFrame([{"join_column": u"🤦🏼‍♂️ L🤦🏼‍♂️R 🤦🏼‍♂️"}])
        left_join_table = pandas.DataFrame([{"join_column": u"🤦🏼‍♂️ L🤦🏼‍♂️R 🤦🏼‍♂️", "other_column": u"🦆🍞🦆"}])
        run_parallel_queries(main_table, left_join_table, left_join_table, pandas)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_parallel_numeric_object(self, duckdb_cursor, pandas):
        main_table = pandas.DataFrame({'join_column': pandas.Series([3], dtype="Int8")})
        left_join_table = pandas.DataFrame(
            {'join_column': pandas.Series([3], dtype="Int8"), 'other_column': pandas.Series([4], dtype="Int8")}
        )
        expected_df = pandas.DataFrame(
            {"join_column": numpy.array([3], dtype=numpy.int8), "other_column": numpy.array([4], dtype=numpy.int8)}
        )
        run_parallel_queries(main_table, left_join_table, expected_df, pandas)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_parallel_timestamp(self, duckdb_cursor, pandas):
        main_table = pandas.DataFrame({'join_column': [pandas.Timestamp('20180310T11:17:54Z')]})
        left_join_table = pandas.DataFrame(
            {
                'join_column': [pandas.Timestamp('20180310T11:17:54Z')],
                'other_column': [pandas.Timestamp('20190310T11:17:54Z')],
            }
        )
        expected_df = pandas.DataFrame(
            {
                "join_column": numpy.array([datetime.datetime(2018, 3, 10, 11, 17, 54)], dtype='datetime64[ns]'),
                "other_column": numpy.array([datetime.datetime(2019, 3, 10, 11, 17, 54)], dtype='datetime64[ns]'),
            }
        )
        run_parallel_queries(main_table, left_join_table, expected_df, pandas)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_parallel_empty(self, duckdb_cursor, pandas):
        df_empty = pandas.DataFrame({'A': []})
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")
        duckdb_conn.execute("PRAGMA verify_parallelism")
        duckdb_conn.register('main_table', df_empty)
        assert duckdb_conn.execute('select * from main_table').fetchall() == []



import pytest
import tempfile

import os

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql import SparkSession as session
from duckdb import connect, InvalidInputException, read_csv
from conftest import NumpyPandas, ArrowPandas, getTimeSeriesData
import pandas._testing as tm
import datetime
import csv


@pytest.fixture
def df(spark):
    simpleData = (
        ("Java", 4000, 5),
        ("Python", 4600, 10),
        ("Scala", 4100, 15),
        ("Scala", 4500, 15),
        ("PHP", 3000, 20),
    )
    columns = ["CourseName", "fee", "discount"]
    dataframe = spark.createDataFrame(data=simpleData, schema=columns)
    yield dataframe


@pytest.fixture(params=[NumpyPandas(), ArrowPandas()])
def pandas_df_ints(request, spark):
    pandas = request.param
    dataframe = pandas.DataFrame({'a': [5, 3, 23, 2], 'b': [45, 234, 234, 2]})
    yield dataframe


@pytest.fixture(params=[NumpyPandas(), ArrowPandas()])
def pandas_df_strings(request, spark):
    pandas = request.param
    dataframe = pandas.DataFrame({'a': ['string1', 'string2', 'string3']})
    yield dataframe


class TestSparkToCSV(object):
    def test_basic_to_csv(self, pandas_df_ints, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")

        df = spark.createDataFrame(pandas_df_ints)

        df.write.csv(temp_file_name, header=False)

        csv_rel = spark.read.csv(temp_file_name, header=False)

        assert df.collect() == csv_rel.collect()

    def test_to_csv_sep(self, pandas_df_ints, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")

        df = spark.createDataFrame(pandas_df_ints)

        df.write.csv(temp_file_name, sep=',', header=False)

        csv_rel = spark.read.csv(temp_file_name, header=False, sep=',')
        assert df.collect() == csv_rel.collect()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_na_rep(self, pandas, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        pandas_df = pandas.DataFrame({'a': [5, None, 23, 2], 'b': [45, 234, 234, 2]})

        df = spark.createDataFrame(pandas_df)

        df.write.csv(temp_file_name, nullValue="test", header=False)

        csv_rel = spark.read.csv(temp_file_name, nullValue="test")
        assert df.collect() == csv_rel.collect()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_header(self, pandas, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        pandas_df = pandas.DataFrame({'a': [5, None, 23, 2], 'b': [45, 234, 234, 2]})

        df = spark.createDataFrame(pandas_df)

        df.write.csv(temp_file_name, header=True)

        csv_rel = spark.read.csv(temp_file_name, header=True)
        assert df.collect() == csv_rel.collect()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_quotechar(self, pandas, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")

        pandas_df = pandas.DataFrame({'a': ["\'a,b,c\'", None, "hello", "bye"], 'b': [45, 234, 234, 2]})

        df = spark.createDataFrame(pandas_df)

        df.write.csv(temp_file_name, quote='\'', sep=',', header=False)

        csv_rel = spark.read.csv(temp_file_name, sep=',', quote='\'')
        assert df.collect() == csv_rel.collect()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_escapechar(self, pandas, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        pandas_df = pandas.DataFrame(
            {
                "c_bool": [True, False],
                "c_float": [1.0, 3.2],
                "c_int": [42, None],
                "c_string": ["a", "b,c"],
            }
        )

        df = spark.createDataFrame(pandas_df)

        df.write.csv(temp_file_name, header=True, quote='"', escape='!')
        csv_rel = spark.read.csv(temp_file_name, quote='"', escape='!', header=True)
        assert df.collect() == csv_rel.collect()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_date_format(self, pandas, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        pandas_df = pandas.DataFrame(getTimeSeriesData())
        dt_index = pandas_df.index
        pandas_df = pandas.DataFrame({"A": dt_index, "B": dt_index.shift(1)}, index=dt_index)

        df = spark.createDataFrame(pandas_df)

        df.write.csv(temp_file_name, dateFormat="%Y%m%d", header=False)

        csv_rel = spark.read.csv(temp_file_name, dateFormat="%Y%m%d")

        assert df.collect() == csv_rel.collect()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_timestamp_format(self, pandas, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        data = [datetime.time(hour=23, minute=1, second=34, microsecond=234345)]
        pandas_df = pandas.DataFrame({'0': pandas.Series(data=data, dtype='object')})

        df = spark.createDataFrame(pandas_df)

        df.write.csv(temp_file_name, timestampFormat='%m/%d/%Y', header=False)

        csv_rel = spark.read.csv(temp_file_name, timestampFormat='%m/%d/%Y')

        assert df.collect() == csv_rel.collect()

    def test_to_csv_quoting_off(self, pandas_df_strings, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        df = spark.createDataFrame(pandas_df_strings)
        df.write.csv(temp_file_name, quoteAll=None, header=False)

        csv_rel = spark.read.csv(temp_file_name)
        assert df.collect() == csv_rel.collect()

    def test_to_csv_quoting_on(self, pandas_df_strings, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        df = spark.createDataFrame(pandas_df_strings)
        df.write.csv(temp_file_name, quoteAll="force", header=False)

        csv_rel = spark.read.csv(temp_file_name)
        assert df.collect() == csv_rel.collect()

    def test_to_csv_quoting_quote_all(self, pandas_df_strings, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        df = spark.createDataFrame(pandas_df_strings)
        df.write.csv(temp_file_name, quoteAll=csv.QUOTE_ALL, header=False)

        csv_rel = spark.read.csv(temp_file_name)
        assert df.collect() == csv_rel.collect()

    def test_to_csv_encoding_incorrect(self, pandas_df_strings, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        df = spark.createDataFrame(pandas_df_strings)
        with pytest.raises(
            InvalidInputException, match="Invalid Input Error: The only supported encoding option is 'UTF8"
        ):
            df.write.csv(temp_file_name, encoding="nope", header=False)

    def test_to_csv_encoding_correct(self, pandas_df_strings, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        df = spark.createDataFrame(pandas_df_strings)
        df.write.csv(temp_file_name, encoding="UTF-8", header=False)
        csv_rel = spark.read.csv(temp_file_name)
        assert df.collect() == csv_rel.collect()

    def test_compression_gzip(self, pandas_df_strings, spark, tmp_path):
        temp_file_name = os.path.join(tmp_path, "temp_file.csv")
        df = spark.createDataFrame(pandas_df_strings)
        df.write.csv(temp_file_name, compression="gzip", header=False)

        # slightly convoluted - pyspark .read.csv does not take a compression argument
        csv_rel = spark.createDataFrame(read_csv(temp_file_name, compression="gzip").df())
        assert df.collect() == csv_rel.collect()



import duckdb
import pytest


@pytest.fixture()
def tbl_table():
    con = duckdb.default_connection
    con.execute("drop table if exists tbl")
    con.execute("create table tbl (i integer)")
    yield
    con.execute('drop table tbl')


class TestRAPIQuery(object):
    @pytest.mark.parametrize('steps', [1, 2, 3, 4])
    def test_query_chain(self, steps):
        con = duckdb.default_connection
        amount = int(1000000)
        rel = None
        for _ in range(steps):
            rel = con.query(f"select i from range({amount}::BIGINT) tbl(i)")
            amount = amount / 10
            rel = rel.query("rel", f"select * from rel limit {amount}")
        result = rel.execute()
        assert len(result.fetchall()) == amount

    @pytest.mark.parametrize('input', [[5, 4, 3], [], [1000]])
    def test_query_table(self, tbl_table, input):
        con = duckdb.default_connection
        rel = con.table("tbl")
        for row in input:
            rel.insert([row])
        # Querying a table relation
        rel = rel.query("x", "select * from x")
        result = rel.execute()
        assert result.fetchall() == [tuple([x]) for x in input]

    def test_query_table_basic(self, tbl_table):
        con = duckdb.default_connection
        rel = con.table("tbl")
        # Querying a table relation
        rel = rel.query("x", "select 5")
        result = rel.execute()
        assert result.fetchall() == [(5,)]

    def test_query_table_qualified(self, duckdb_cursor):
        con = duckdb.default_connection
        con.execute("create schema fff")

        # Create table in fff schema
        con.execute("create table fff.t2 as select 1 as t")
        assert con.table("fff.t2").fetchall() == [(1,)]

    def test_query_insert_into_relation(self, tbl_table):
        con = duckdb.default_connection
        rel = con.query("select i from range(1000) tbl(i)")
        # Can't insert into this, not a table relation
        with pytest.raises(duckdb.InvalidInputException):
            rel.insert([5])

    def test_query_non_select(self, duckdb_cursor):
        rel = duckdb_cursor.query("select [1,2,3,4]")
        rel.query("relation", "create table tbl as select * from relation")

        result = duckdb_cursor.execute("select * from tbl").fetchall()
        assert result == [([1, 2, 3, 4],)]

    def test_query_non_select_fail(self, duckdb_cursor):
        rel = duckdb_cursor.query("select [1,2,3,4]")
        duckdb_cursor.execute("create table tbl as select range(10)")
        # Table already exists
        with pytest.raises(duckdb.CatalogException):
            rel.query("relation", "create table tbl as select * from relation")

        # View referenced does not exist
        with pytest.raises(duckdb.CatalogException):
            rel.query("relation", "create table tbl as select * from not_a_valid_view")

    def test_query_table_unrelated(self, tbl_table):
        con = duckdb.default_connection
        rel = con.table("tbl")
        # Querying a table relation
        rel = rel.query("x", "select 5")
        result = rel.execute()
        assert result.fetchall() == [(5,)]

    def test_query_non_select_result(self, duckdb_cursor):
        with pytest.raises(duckdb.ParserException, match="syntax error"):
            duckdb_cursor.query('selec 42')

        res = duckdb_cursor.query('explain select 42').fetchall()
        assert len(res) > 0

        res = duckdb_cursor.query('describe select 42::INT AS column_name').fetchall()
        assert res[0][0] == 'column_name'

        res = duckdb_cursor.query('create or replace table tbl_non_select_result(i integer)')
        assert res is None

        res = duckdb_cursor.query('insert into tbl_non_select_result values (42)')
        assert res is None

        res = duckdb_cursor.query('insert into tbl_non_select_result values (84) returning *').fetchall()
        assert res == [(84,)]

        res = duckdb_cursor.query('select * from tbl_non_select_result').fetchall()
        assert res == [(42,), (84,)]

        res = duckdb_cursor.query('insert into tbl_non_select_result select * from range(10000) returning *').fetchall()
        assert len(res) == 10000

        res = duckdb_cursor.query('show tables').fetchall()
        assert len(res) > 0

        res = duckdb_cursor.query('drop table tbl_non_select_result')
        assert res is None

    def test_replacement_scan_recursion(self, duckdb_cursor):
        depth_limit = 1000
        import sys

        if sys.platform.startswith('win'):
            # With the default we reach a stack overflow in the CI
            depth_limit = 250
        duckdb_cursor.execute(f"SET max_expression_depth TO {depth_limit}")
        rel = duckdb_cursor.sql('select 42')
        rel = duckdb_cursor.sql('select * from rel')
        with pytest.raises(duckdb.BinderException, match=f'Max expression depth limit of {depth_limit} exceeded'):
            duckdb_cursor.sql('select * from rel')



import numpy
import pandas
from decimal import *


class TestDecimal(object):
    def test_decimal(self, duckdb_cursor):
        duckdb_cursor.execute(
            'SELECT 1.2::DECIMAL(4,1), 100.3::DECIMAL(9,1), 320938.4298::DECIMAL(18,4), 49082094824.904820482094::DECIMAL(30,12), NULL::DECIMAL'
        )
        result = duckdb_cursor.fetchall()
        assert result == [
            (Decimal('1.2'), Decimal('100.3'), Decimal('320938.4298'), Decimal('49082094824.904820482094'), None)
        ]

    def test_decimal_numpy(self, duckdb_cursor):
        duckdb_cursor.execute(
            'SELECT 1.2::DECIMAL(4,1) AS a, 100.3::DECIMAL(9,1) AS b, 320938.4298::DECIMAL(18,4) AS c, 49082094824.904820482094::DECIMAL(30,12) AS d'
        )
        result = duckdb_cursor.fetchnumpy()
        assert result == {
            'a': numpy.array([1.2]),
            'b': numpy.array([100.3]),
            'c': numpy.array([320938.4298]),
            'd': numpy.array([49082094824.904820482094]),
        }



from ... import DuckDBPyConnection
from ...typing import DuckDBPyType
from typing import Any

class NullValue(Value):
    def __init__(self) -> None: ...
    def __repr__(self) -> str: ...

class BooleanValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class UnsignedBinaryValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class UnsignedShortValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class UnsignedIntegerValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class UnsignedLongValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class BinaryValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class ShortValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class IntegerValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class LongValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class HugeIntegerValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class FloatValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class DoubleValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class DecimalValue(Value):
        def __init__(self, object: Any, width: int, scale: int) -> None: ...
        def __repr__(self) -> str: ...
    
class StringValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class UUIDValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class BitValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class BlobValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class DateValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class IntervalValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class TimestampValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class TimestampSecondValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class TimestampMilisecondValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class TimestampNanosecondValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class TimestampTimeZoneValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class TimeValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...

class TimeTimeZoneValue(Value):
    def __init__(self, object: Any) -> None: ...
    def __repr__(self) -> str: ...


class Value:
    def __init__(self, object: Any, type: DuckDBPyType) -> None: ...
    def __repr__(self) -> str: ...



from pytest import importorskip

importorskip('pyarrow')

import duckdb
from pyarrow import scalar, string, large_string, list_, int32, types


def test_nested(duckdb_cursor):
    res = run(duckdb_cursor, 'select 42::UNION(name VARCHAR, attr UNION(age INT, veteran BOOL)) as res')
    assert types.is_union(res.type)
    assert res.value.value == scalar(42, type=int32())


def test_union_contains_nested_data(duckdb_cursor):
    _ = importorskip("pyarrow", minversion="11")
    res = run(duckdb_cursor, "select ['hello']::UNION(first_name VARCHAR, middle_names VARCHAR[]) as res")
    assert types.is_union(res.type)
    assert res.value == scalar(['hello'], type=list_(string()))


def test_unions_inside_lists_structs_maps(duckdb_cursor):
    res = run(duckdb_cursor, "select [union_value(name := 'Frank')] as res")
    assert types.is_list(res.type)
    assert types.is_union(res.type.value_type)
    assert res[0].value == scalar('Frank', type=string())


def test_unions_with_struct(duckdb_cursor):
    duckdb_cursor.execute(
        """
		CREATE TABLE tbl (a UNION(a STRUCT(a INT, b BOOL)))
	"""
    )
    duckdb_cursor.execute(
        """
		INSERT INTO tbl VALUES ({'a': 42, 'b': true})
	"""
    )

    rel = duckdb_cursor.table('tbl')
    arrow = rel.arrow()

    duckdb_cursor.execute("create table other as select * from arrow")
    rel2 = duckdb_cursor.table('other')
    res = rel2.fetchall()
    assert res == [({'a': 42, 'b': True},)]


def run(conn, query):
    return conn.sql(query).arrow().columns[0][0]



from concurrent.futures import ThreadPoolExecutor
import duckdb
import pytest

pyarrow = pytest.importorskip('pyarrow')


def f(cur, i, data):
    cur.execute(f"create table t_{i} as select * from data")
    return cur.execute(f"select * from t_{i}").arrow()


def test_6584():
    pool = ThreadPoolExecutor(max_workers=2)
    data = pyarrow.Table.from_pydict({"a": [1, 2, 3]})
    c = duckdb.connect()
    futures = []
    for i in range(2):
        fut = pool.submit(f, c.cursor(), i, data)
        futures.append(fut)

    for fut in futures:
        arrow_res = fut.result()
        assert data.equals(arrow_res)



import pytest
import duckdb


class TestExplain(object):
    def test_explain_basic(self, duckdb_cursor):
        res = duckdb_cursor.sql('select 42').explain()
        assert isinstance(res, str)

    def test_explain_standard(self, duckdb_cursor):
        res = duckdb_cursor.sql('select 42').explain('standard')
        assert isinstance(res, str)

        res = duckdb_cursor.sql('select 42').explain('STANDARD')
        assert isinstance(res, str)

        res = duckdb_cursor.sql('select 42').explain(duckdb.STANDARD)
        assert isinstance(res, str)

        res = duckdb_cursor.sql('select 42').explain(duckdb.ExplainType.STANDARD)
        assert isinstance(res, str)

        res = duckdb_cursor.sql('select 42').explain(0)
        assert isinstance(res, str)

    def test_explain_analyze(self, duckdb_cursor):
        res = duckdb_cursor.sql('select 42').explain('analyze')
        assert isinstance(res, str)

        res = duckdb_cursor.sql('select 42').explain('ANALYZE')
        assert isinstance(res, str)

        res = duckdb_cursor.sql('select 42').explain(duckdb.ANALYZE)
        assert isinstance(res, str)

        res = duckdb_cursor.sql('select 42').explain(duckdb.ExplainType.ANALYZE)
        assert isinstance(res, str)

        res = duckdb_cursor.sql('select 42').explain(1)
        assert isinstance(res, str)

    def test_explain_df(self, duckdb_cursor):
        pd = pytest.importorskip("pandas")
        df = pd.DataFrame({'a': [42]})
        res = duckdb_cursor.sql('select * from df').explain('ANALYZE')
        assert isinstance(res, str)



import duckdb
import pandas as pd


class TestConnectionTransaction(object):
    def test_transaction(self, duckdb_cursor):
        con = duckdb.connect()
        con.execute('create table t (i integer)')
        con.execute('insert into t values (1)')

        con.begin()
        con.execute('insert into t values (1)')
        assert con.execute('select count (*) from t').fetchone()[0] == 2
        con.rollback()
        assert con.execute('select count (*) from t').fetchone()[0] == 1
        con.begin()
        con.execute('insert into t values (1)')
        assert con.execute('select count (*) from t').fetchone()[0] == 2
        con.commit()
        assert con.execute('select count (*) from t').fetchone()[0] == 2



import pandas
import numpy as np
import datetime
import duckdb
import pytest
from conftest import NumpyPandas, ArrowPandas


class TestCaseAlias(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_case_alias(self, duckdb_cursor, pandas):
        import numpy as np
        import datetime
        import duckdb

        con = duckdb.connect(':memory:')

        df = pandas.DataFrame([{"COL1": "val1", "CoL2": 1.05}, {"COL1": "val3", "CoL2": 17}])

        r1 = con.from_df(df).query('df', 'select * from df').df()
        assert r1["COL1"][0] == "val1"
        assert r1["COL1"][1] == "val3"
        assert r1["CoL2"][0] == 1.05
        assert r1["CoL2"][1] == 17

        r2 = con.from_df(df).query('df', 'select COL1, COL2 from df').df()
        assert r2["COL1"][0] == "val1"
        assert r2["COL1"][1] == "val3"
        assert r2["CoL2"][0] == 1.05
        assert r2["CoL2"][1] == 17

        r3 = con.from_df(df).query('df', 'select COL1, COL2 from df ORDER BY COL1').df()
        assert r3["COL1"][0] == "val1"
        assert r3["COL1"][1] == "val3"
        assert r3["CoL2"][0] == 1.05
        assert r3["CoL2"][1] == 17

        r4 = con.from_df(df).query('df', 'select COL1, COL2 from df GROUP BY COL1, COL2 ORDER BY COL1').df()
        assert r4["COL1"][0] == "val1"
        assert r4["COL1"][1] == "val3"
        assert r4["CoL2"][0] == 1.05
        assert r4["CoL2"][1] == 17



import duckdb
import pytest


torch = pytest.importorskip('torch')


def test_pytorch():
    con = duckdb.connect()

    con.execute("create table t( a integer, b integer)")
    con.execute("insert into t values (1,2), (3,4)")

    # Test from connection
    duck_torch = con.execute("select * from t").torch()
    duck_numpy = con.sql("select * from t").fetchnumpy()
    torch.equal(duck_torch['a'], torch.tensor(duck_numpy['a']))
    torch.equal(duck_torch['b'], torch.tensor(duck_numpy['b']))

    # Test from relation
    duck_torch = con.sql("select * from t").torch()
    torch.equal(duck_torch['a'], torch.tensor(duck_numpy['a']))
    torch.equal(duck_torch['b'], torch.tensor(duck_numpy['b']))

    # Test all Numeric Types
    numeric_types = ['TINYINT', 'SMALLINT', 'BIGINT', 'HUGEINT', 'FLOAT', 'DOUBLE', 'DECIMAL(4,1)', 'UTINYINT']

    for supported_type in numeric_types:
        con = duckdb.connect()
        con.execute(f"create table t( a {supported_type} , b {supported_type})")
        con.execute("insert into t values (1,2), (3,4)")
        duck_torch = con.sql("select * from t").torch()
        duck_numpy = con.sql("select * from t").fetchnumpy()
        torch.equal(duck_torch['a'], torch.tensor(duck_numpy['a']))
        torch.equal(duck_torch['b'], torch.tensor(duck_numpy['b']))

    with pytest.raises(TypeError, match="can't convert"):
        con = duckdb.connect()
        con.execute(f"create table t( a UINTEGER)")
        duck_torch = con.sql("select * from t").torch()
